# ruff: noqa: F401, F403

from ._lib import __version__
from .casting import st
from .geodataframe import *
from .geoexpr import *
from .geolazyframe import *
from .geometry import *
from .geoseries import *
from .input import *
from .parsing import *
from .selectors import *
from .sugar import *



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, cast

import polars as pl
from polars._utils.parse import parse_into_expression
from polars._utils.wrap import wrap_expr
from polars.plugins import register_plugin_function

if TYPE_CHECKING:
    from polars._typing import IntoExprColumn

    from polars_st.geoexpr import GeoExpr
    from polars_st.typing import IntoIntegerExpr


__all__ = [
    "circularstring",
    "circularstring",
    "from_ewkt",
    "from_geojson",
    "from_shapely",
    "from_wkb",
    "from_wkt",
    "linestring",
    "multilinestring",
    "multipoint",
    "point",
    "polygon",
    "rectangle",
]


def point(coords: IntoExprColumn, srid: IntoIntegerExpr = 0) -> GeoExpr:
    """Create Point geometries from coordinates.

    Examples:
        >>> df = pl.DataFrame({
        ...     "coords": [
        ...          [0, 1],
        ...          [2, 3],
        ...     ]
        ... })
        >>> df = df.select(geometry=st.point("coords"))
        >>> df.st.to_wkt()
        shape: (2, 1)
        ┌─────────────┐
        │ geometry    │
        │ ---         │
        │ str         │
        ╞═════════════╡
        │ POINT (0 1) │
        │ POINT (2 3) │
        └─────────────┘

        >>> df = pl.DataFrame({
        ...     "x": [0, 1],
        ...     "y": [0, 2],
        ...     "z": [0, 3],
        ... })
        >>> df = df.select(geometry=st.point(pl.concat_arr("x", "y", "z")))
        >>> df.st.to_wkt()
        shape: (2, 1)
        ┌─────────────────┐
        │ geometry        │
        │ ---             │
        │ str             │
        ╞═════════════════╡
        │ POINT Z (0 0 0) │
        │ POINT Z (1 2 3) │
        └─────────────────┘
    """
    return register_plugin_function(
        plugin_path=Path(__file__).parent,
        function_name="point",
        args=[coords, srid],
        is_elementwise=True,
    ).pipe(lambda e: cast("GeoExpr", e))


def multipoint(coords: IntoExprColumn, srid: IntoIntegerExpr = 0) -> GeoExpr:
    """Create MultiPoint geometries from list of coordinates.

    Examples:
        >>> df = pl.DataFrame({
        ...     "coords": [
        ...          [[0, 1],[2, 3]],
        ...          [[4, 5],[6, 7]],
        ...     ]
        ... })
        >>> df = df.select(geometry=st.multipoint("coords"))
        >>> df.st.to_wkt()
        shape: (2, 1)
        ┌───────────────────────────┐
        │ geometry                  │
        │ ---                       │
        │ str                       │
        ╞═══════════════════════════╡
        │ MULTIPOINT ((0 1), (2 3)) │
        │ MULTIPOINT ((4 5), (6 7)) │
        └───────────────────────────┘
    """
    return register_plugin_function(
        plugin_path=Path(__file__).parent,
        function_name="multipoint",
        args=[coords, srid],
        is_elementwise=True,
    ).pipe(lambda e: cast("GeoExpr", e))


def linestring(coords: IntoExprColumn, srid: IntoIntegerExpr = 0) -> GeoExpr:
    """Create LineString geometries from lists of coordinates.

    Examples:
        >>> df = pl.DataFrame({
        ...     "coords": [
        ...         [[0, 1], [2, 3], [4, 5]]
        ...     ],
        ... })
        >>> df = df.select(geometry=st.linestring("coords"))
        >>> df.st.to_wkt()
        shape: (1, 1)
        ┌────────────────────────────┐
        │ geometry                   │
        │ ---                        │
        │ str                        │
        ╞════════════════════════════╡
        │ LINESTRING (0 1, 2 3, 4 5) │
        └────────────────────────────┘

        >>> df = pl.DataFrame({
        ...     "idx": [0, 0, 1, 1],
        ...     "x": [0, 1, 3, 5],
        ...     "y": [0, 2, 4, 6],
        ... })
        >>> df = df.group_by("idx").agg(coords=pl.concat_list("x", "y"))
        >>> df = df.select("idx", geometry=st.linestring("coords"))
        >>> df.sort("idx").st.to_wkt()
        shape: (2, 2)
        ┌─────┬───────────────────────┐
        │ idx ┆ geometry              │
        │ --- ┆ ---                   │
        │ i64 ┆ str                   │
        ╞═════╪═══════════════════════╡
        │ 0   ┆ LINESTRING (0 0, 1 2) │
        │ 1   ┆ LINESTRING (3 4, 5 6) │
        └─────┴───────────────────────┘
    """
    return register_plugin_function(
        plugin_path=Path(__file__).parent,
        function_name="linestring",
        args=[coords, srid],
        is_elementwise=True,
    ).pipe(lambda e: cast("GeoExpr", e))


def circularstring(coords: IntoExprColumn, srid: IntoIntegerExpr = 0) -> GeoExpr:
    """Create CircularString geometries from lists of coordinates.

    Examples:
        >>> df = pl.DataFrame({
        ...     "coords": [
        ...         [[0, 1], [2, 3], [4, 5]]
        ...     ],
        ... })
        >>> df = df.select(geometry=st.circularstring("coords"))
        >>> df.st.to_wkt()
        shape: (1, 1)
        ┌────────────────────────────────┐
        │ geometry                       │
        │ ---                            │
        │ str                            │
        ╞════════════════════════════════╡
        │ CIRCULARSTRING (0 1, 2 3, 4 5) │
        └────────────────────────────────┘
    """
    return register_plugin_function(
        plugin_path=Path(__file__).parent,
        function_name="circularstring",
        args=[coords, srid],
        is_elementwise=True,
    ).pipe(lambda e: cast("GeoExpr", e))


def multilinestring(coords: IntoExprColumn, srid: IntoIntegerExpr = 0) -> GeoExpr:
    """Create MultiLineString geometries from lists of lists of coordinates.

    Examples:
        >>> df = pl.DataFrame({
        ...     "coords": [
        ...         [[[1, 2], [3, 4]],[[5, 6], [7, 8]]]
        ...     ]
        ... })
        >>> df = df.select(geometry=st.multilinestring("coords"))
        >>> df.st.to_wkt()
        shape: (1, 1)
        ┌─────────────────────────────────┐
        │ geometry                        │
        │ ---                             │
        │ str                             │
        ╞═════════════════════════════════╡
        │ MULTILINESTRING ((1 2, 3 4), (… │
        └─────────────────────────────────┘
    """
    return register_plugin_function(
        plugin_path=Path(__file__).parent,
        function_name="multilinestring",
        args=[coords, srid],
        is_elementwise=True,
    ).pipe(lambda e: cast("GeoExpr", e))


def polygon(coords: IntoExprColumn, srid: IntoIntegerExpr = 0) -> GeoExpr:
    """Create Polygon geometries from lists of lists of coordinates.

    Examples:
        >>> df = pl.DataFrame({
        ...     "coords": [
        ...         [[[0, 0], [2, 4], [4, 0], [0, 0]]]
        ...     ]
        ... })
        >>> df = df.select(geometry=st.polygon("coords"))
        >>> df.st.to_wkt()
        shape: (1, 1)
        ┌────────────────────────────────┐
        │ geometry                       │
        │ ---                            │
        │ str                            │
        ╞════════════════════════════════╡
        │ POLYGON ((0 0, 2 4, 4 0, 0 0)) │
        └────────────────────────────────┘
    """
    return register_plugin_function(
        plugin_path=Path(__file__).parent,
        function_name="polygon",
        args=[coords, srid],
        is_elementwise=True,
    ).pipe(lambda e: cast("GeoExpr", e))


def rectangle(bounds: IntoExprColumn, srid: IntoIntegerExpr = 0) -> GeoExpr:
    """Create Polygon geometries from bounds.

    Examples:
        >>> df = pl.DataFrame({
        ...     "bounds": [
        ...         [0.0, 0.0, 1.0, 2.0],
        ...         [5.0, 6.0, 7.0, 8.0],
        ...     ]
        ... })
        >>> df = df.select(geometry=st.rectangle("bounds"))
        >>> df.st.to_wkt()
        shape: (2, 1)
        ┌─────────────────────────────────┐
        │ geometry                        │
        │ ---                             │
        │ str                             │
        ╞═════════════════════════════════╡
        │ POLYGON ((0 0, 1 0, 1 2, 0 2, … │
        │ POLYGON ((5 6, 7 6, 7 8, 5 8, … │
        └─────────────────────────────────┘
    """
    return register_plugin_function(
        plugin_path=Path(__file__).parent,
        function_name="rectangle",
        args=[bounds, srid],
        is_elementwise=True,
    ).pipe(lambda e: cast("GeoExpr", e))


def from_wkb(expr: IntoExprColumn) -> GeoExpr:
    """Parse geometries from Well-Known Binary (WKB) representation.

    Examples:
        >>> df = pl.read_database(
        ...     query="SELECT ST_AsEWKB(geom) AS geometry FROM test_data",
        ...     connection=user_conn,
        ... ) # doctest: +SKIP
        >>> gdf = df.select(st.from_wkb("geometry")) # doctest: +SKIP
    """
    return register_plugin_function(
        plugin_path=Path(__file__).parent,
        function_name="from_wkb",
        args=[expr],
        is_elementwise=True,
    ).pipe(lambda e: cast("GeoExpr", e))


def from_wkt(expr: IntoExprColumn) -> GeoExpr:
    """Parse geometries from Well-Known Text (WKT) representation.

    Examples:
        >>> df = pl.Series("geometry", [
        ...     "POINT(0 0)",
        ...     "POINT(1 2)",
        ... ]).to_frame()
        >>> gdf = df.select(st.from_wkt("geometry"))
    """
    return register_plugin_function(
        plugin_path=Path(__file__).parent,
        function_name="from_wkt",
        args=[expr],
        is_elementwise=True,
    ).pipe(lambda e: cast("GeoExpr", e))


def from_ewkt(expr: IntoExprColumn) -> GeoExpr:
    """Parse geometries from Extended Well-Known Text (EWKT) representation.

    Examples:
        >>> df = pl.Series("geometry", [
        ...     "SRID=4326;POINT(0 0)",
        ...     "SRID=3857;POINT(1 2)",
        ... ]).to_frame()
        >>> gdf = df.select(st.from_ewkt("geometry"))
    """
    return register_plugin_function(
        plugin_path=Path(__file__).parent,
        function_name="from_ewkt",
        args=[expr],
        is_elementwise=True,
    ).pipe(lambda e: cast("GeoExpr", e))


def from_geojson(expr: IntoExprColumn) -> GeoExpr:
    """Parse geometries from GeoJSON representation.

    Examples:
        >>> df = pl.Series("geometry", [
        ...     '{"type": "Point", "coordinates": [0, 0]}',
        ...     '{"type": "Point", "coordinates": [1, 2]}',
        ... ]).to_frame()
        >>> gdf = df.select(st.from_geojson("geometry"))
        >>> gdf.st.to_wkt()
        shape: (2, 1)
        ┌─────────────┐
        │ geometry    │
        │ ---         │
        │ str         │
        ╞═════════════╡
        │ POINT (0 0) │
        │ POINT (1 2) │
        └─────────────┘
    """
    return register_plugin_function(
        plugin_path=Path(__file__).parent,
        function_name="from_geojson",
        args=[expr],
        is_elementwise=True,
    ).pipe(lambda e: cast("GeoExpr", e))


def from_shapely(expr: IntoExprColumn) -> GeoExpr:
    """Parse geometries from shapely objects.

    Examples:
        >>> import shapely
        >>> df = pl.Series("geometry", [
        ...     shapely.Point(0, 0),
        ...     shapely.Point(1, 2),
        ... ], dtype=pl.Object).to_frame()
        >>> df.select(st.from_shapely("geometry")).st.to_wkt()
        shape: (2, 1)
        ┌─────────────┐
        │ geometry    │
        │ ---         │
        │ str         │
        ╞═════════════╡
        │ POINT (0 0) │
        │ POINT (1 2) │
        └─────────────┘
    """
    import shapely

    expr = wrap_expr(parse_into_expression(expr))
    res = expr.map_batches(
        lambda s: pl.Series(
            s.name,
            list(shapely.to_wkb(s.to_numpy(), include_srid=True)),
            pl.Binary,
        ),
        return_dtype=pl.Binary,
        is_elementwise=True,
    )
    return cast("GeoExpr", res)



import polars as pl
from polars._plr import PySeries

__version__: str

def get_crs_from_code(srid: int) -> str | None: ...
def get_crs_authority(definition: str) -> tuple[str, str] | None: ...
def to_python_dict(series: PySeries) -> pl.Series: ...



from __future__ import annotations

from typing import TYPE_CHECKING, overload

if TYPE_CHECKING:
    from polars import DataFrame, Expr, LazyFrame, Series

    from polars_st.geodataframe import GeoDataFrameNameSpace
    from polars_st.geoexpr import GeoExprNameSpace
    from polars_st.geolazyframe import GeoLazyFrameNameSpace
    from polars_st.geoseries import GeoSeriesNameSpace


@overload
def st(d: Expr) -> GeoExprNameSpace: ...


@overload
def st(d: Series) -> GeoSeriesNameSpace: ...


@overload
def st(d: DataFrame) -> GeoDataFrameNameSpace: ...


@overload
def st(d: LazyFrame) -> GeoLazyFrameNameSpace: ...


def st(
    d: Expr | Series | DataFrame | LazyFrame,
) -> GeoExprNameSpace | GeoSeriesNameSpace | GeoDataFrameNameSpace | GeoLazyFrameNameSpace:
    return d.st  # type: ignore  # noqa: PGH003



from __future__ import annotations

from typing import TYPE_CHECKING, cast

import polars as pl

if TYPE_CHECKING:
    from collections.abc import Iterable

    from polars_st.geoexpr import GeoExpr


__all__ = [
    "element",
    "geom",
]


def geom(name: str | Iterable[str] = "geometry", *more_names: str) -> GeoExpr:
    """Create a geometry column expression.

    Alias for
        [`polars.col`](https://docs.pola.rs/api/python/stable/reference/expressions/col.html)
    with type hints for the `st` namespace.

    Args:
        name: The name or datatype of the geometry column(s) to represent. Accepts regular
            expression input. Regular expressions should start with `^` and end with `$`.
        more_names: Additional names or datatypes of columns to represent, specified as positional
            arguments.

    Examples:
        Pass a single column name to represent that column:

        >>> gdf = st.GeoSeries("my_geom", ["POINT(0 0)"]).to_frame()
        >>> gdf.select(st.geom("my_geom").st.to_wkt())
        shape: (1, 1)
        ┌─────────────┐
        │ my_geom     │
        │ ---         │
        │ str         │
        ╞═════════════╡
        │ POINT (0 0) │
        └─────────────┘

        Call `geom` without a column name to use the default:

        >>> gdf = st.GeoDataFrame([
        ...     "POINT(0 0)",
        ...     "POINT(1 2)",
        ... ])
        >>> gdf.select(st.geom().st.union_all().st.to_wkt())
        shape: (1, 1)
        ┌───────────────────────────┐
        │ geometry                  │
        │ ---                       │
        │ str                       │
        ╞═══════════════════════════╡
        │ MULTIPOINT ((0 0), (1 2)) │
        └───────────────────────────┘
    """
    return cast("GeoExpr", pl.col(name, *more_names))


def element() -> GeoExpr:
    """Alias for [`polars.element`](https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.element.html).

    Examples:
        >>> gdf = st.GeoDataFrame([
        ...     "MULTIPOINT ((0 0), (1 2))"
        ... ])
        >>> gdf.select(st.parts().list.eval(st.element().st.to_wkt()))
        shape: (1, 1)
        ┌────────────────────────────────┐
        │ geometry                       │
        │ ---                            │
        │ list[str]                      │
        ╞════════════════════════════════╡
        │ ["POINT (0 0)", "POINT (1 2)"] │
        └────────────────────────────────┘
    """
    return cast("GeoExpr", pl.element())



from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING, Any, Literal, ParamSpec, cast

import polars as pl
from polars.api import register_series_namespace

from polars_st.casting import st
from polars_st.parsing import (
    circularstring,
    from_ewkt,
    from_geojson,
    from_shapely,
    from_wkt,
    linestring,
    multilinestring,
    multipoint,
    point,
    polygon,
)
from polars_st.utils.internal import is_empty_method

if TYPE_CHECKING:
    from collections.abc import Sequence

    import altair as alt
    import geopandas as gpd
    from altair.vegalite.v6.schema._config import MarkConfigKwds
    from lonboard import Map
    from lonboard.types.layer import (
        PathLayerKwargs,
        PolygonLayerKwargs,
        ScatterplotLayerKwargs,
    )
    from lonboard.types.map import MapKwargs
    from polars._typing import PolarsDataType
    from typing_extensions import Unpack

    from polars_st.geometry import CoordinateType, GeometryType
    from polars_st.typing import (
        ArrayLike,
        IntoExprColumn,
        IntoGeoExprColumn,
        IntoIntegerExpr,
        IntoNumericExpr,
    )

    P = ParamSpec("P")


__all__ = [
    "GeoSeries",
    "GeoSeriesNameSpace",
]


class GeoSeriesMeta(type):
    def __instancecheck__(cls, instance: Any) -> bool:
        # The GeoSeries constructor doesn't return an instance of GeoSeries but an
        # instance of pl.DataFrame. This design decision is made because Polars doesn't
        # support subclassing of its code datatypes. In order to prevent misuse,
        # instance checks are forbidden.
        msg = "instance check on abstract class GeoSeries is not allowed"
        raise TypeError(msg)


class GeoSeries(pl.Series, metaclass=GeoSeriesMeta):
    @property
    def st(self) -> GeoSeriesNameSpace:
        return GeoSeriesNameSpace(self)

    def __new__(  # noqa: C901, PLR0912, PLR0915, PYI034
        cls,
        name: str | ArrayLike | None = None,
        values: ArrayLike | None = None,
        dtype: PolarsDataType | None = None,
        *,
        strict: bool = True,
        nan_to_null: bool = False,
        geometry_format: Literal[
            "wkb",
            "wkt",
            "ewkt",
            "geojson",
            "shapely",
            "point",
            "multipoint",
            "linestring",
            "circularstring",
            "multilinestring",
            "polygon",
        ]
        | None = None,
    ) -> GeoSeries:
        s = pl.Series(name, values, dtype, strict=strict, nan_to_null=nan_to_null)
        if s.name == "" and not (isinstance(name, str) and name == ""):
            s = s.rename("geometry")
        if len(s) == 0 or s.dtype == pl.Null:
            return cast("GeoSeries", s.cast(pl.Binary))
        if geometry_format is None:
            match s.dtype:
                case pl.Binary:
                    geometry_format = "wkb"
                case pl.String:
                    first_value: str | None = s[cast("int", s.is_not_null().arg_max())]
                    if first_value is None:
                        return cast("GeoSeries", s.cast(pl.Binary))
                    if first_value.startswith("{"):
                        geometry_format = "geojson"
                    elif first_value.startswith("SRID="):
                        geometry_format = "ewkt"
                    else:
                        geometry_format = "wkt"
                case pl.Object:
                    geometry_format = "shapely"
                case pl.List | pl.Array:
                    inner = s.dtype.inner
                    if inner.is_numeric():
                        geometry_format = "point"
                    elif inner in (pl.List, pl.Array):
                        inner = inner.inner
                        if inner.is_numeric():
                            geometry_format = "linestring"
                        elif inner in (pl.List, pl.Array):
                            inner = inner.inner
                            if inner.is_numeric():
                                geometry_format = "polygon"

        match geometry_format:
            case None:
                msg = f"Couldn't infer geometry format from dtype {s.dtype}"
                raise ValueError(msg)
            case "wkb":
                result = s
            case "wkt":
                result = pl.select(from_wkt(s)).to_series()
            case "ewkt":
                result = pl.select(from_ewkt(s)).to_series()
            case "geojson":
                result = pl.select(from_geojson(s)).to_series()
            case "shapely":
                result = pl.select(from_shapely(s)).to_series()
            case "point":
                result = pl.select(point(s)).to_series()
            case "multipoint":
                result = pl.select(multipoint(s)).to_series()
            case "linestring":
                result = pl.select(linestring(s)).to_series()
            case "circularstring":
                result = pl.select(circularstring(s)).to_series()
            case "multilinestring":
                result = pl.select(multilinestring(s)).to_series()
            case "polygon":
                result = pl.select(polygon(s)).to_series()
        return cast("GeoSeries", result)

    def __init__(
        self,
        name: str | ArrayLike | None = None,
        values: ArrayLike | None = None,
        dtype: PolarsDataType | None = None,
        *,
        strict: bool = True,
        nan_to_null: bool = False,
        geometry_format: Literal[
            "wkb",
            "wkt",
            "ewkt",
            "geojson",
            "shapely",
            "point",
            "multipoint",
            "linestring",
            "circularstring",
            "multilinestring",
            "polygon",
        ]
        | None = None,
    ) -> None:
        """Create a new GeoSeries.

        `GeoSeries` is used as an alias for `pl.Series` with type annotations added for the
        [`st`][polars_st.GeoSeries.st] namespace, and an overriden constructor which will parse
        the values into binary EWKB format.

        You can create a GeoSeries from a list of coordinate arrays, WKB, WKT, EWKT or GeoJSON
            strings, or Shapely objects. If `geometry_format` is not set, the geometries will be
            created by infering the correct deserialization operation from its datatype.

        See [`pl.Series`](https://docs.pola.rs/api/python/stable/reference/series/index.html)
        for parameters documentation.

        !!! note

            Because Polars doesn't support subclassing of their types, calling this constructor will
            **NOT** create an instance of `GeoSeries`, but an instance of `pl.Series`.

            As a result, instance checks are not permitted on this class to prevent misuse:
            ```pycon
            >>> s = st.GeoSeries(["POINT(0 0)"])
            >>> type(s)
            <class 'polars.series.series.Series'>
            >>> isinstance(s, st.GeoSeries)
            Traceback (most recent call last):
            ...
            TypeError: instance check on abstract class GeoSeries is not allowed
            ```

        Examples:
            >>> gs = st.GeoSeries([
            ...     "POINT(0 0)",
            ...     "POINT(1 2)",
            ... ])
            >>> gs2 = st.GeoSeries([
            ...     [0, 0],
            ...     [1, 2],
            ... ], geometry_format="point")
            >>> gs.equals(gs2)
            True

            >>> import shapely
            >>> gs = st.GeoSeries([
            ...     shapely.Point(0, 0),
            ...     shapely.Point(1, 2),
            ... ])
            >>> gs2 = st.GeoSeries([
            ...     '{"type": "Point", "coordinates": [0, 0]}',
            ...     '{"type": "Point", "coordinates": [1, 2]}',
            ... ])
            >>> gs.equals(gs2)
            True
        """
        ...


def dispatch(func):  # noqa: ANN001, ANN202 to preserve pylance type hints
    assert is_empty_method(func)  # noqa: S101

    @wraps(func)
    def wrapper(self: GeoSeriesNameSpace, *args: P.args, **kwargs: P.kwargs) -> pl.Series:
        f = getattr(getattr(pl.col(self._series.name), "st"), func.__name__)  # noqa: B009
        return self._series.to_frame().select_seq(f(*args, **kwargs)).to_series()

    return wrapper


@register_series_namespace("st")
class GeoSeriesNameSpace:
    def __init__(self, series: pl.Series) -> None:
        self._series = cast("GeoSeries", series)

    @dispatch
    def geometry_type(self) -> pl.Series:
        """See [`GeoExprNameSpace.geometry_type`][polars_st.GeoExprNameSpace.geometry_type]."""
        ...

    @dispatch
    def dimension(self) -> pl.Series:
        """See [`GeoExprNameSpace.dimension`][polars_st.GeoExprNameSpace.dimension]."""
        ...

    @dispatch
    def coordinate_dimension(self) -> pl.Series:
        """See [`GeoExprNameSpace.coordinate_dimension`][polars_st.GeoExprNameSpace.coordinate_dimension]."""  # noqa: E501
        ...

    @dispatch
    def coordinate_type(self) -> pl.Series:
        """See [`GeoExprNameSpace.coordinate_type`][polars_st.GeoExprNameSpace.coordinate_type]."""
        ...

    @dispatch
    def area(self) -> pl.Series:
        """See [`GeoExprNameSpace.area`][polars_st.GeoExprNameSpace.area]."""
        ...

    @dispatch
    def bounds(self) -> pl.Series:
        """See [`GeoExprNameSpace.bounds`][polars_st.GeoExprNameSpace.bounds]."""
        ...

    @dispatch
    def length(self) -> pl.Series:
        """See [`GeoExprNameSpace.length`][polars_st.GeoExprNameSpace.length]."""
        ...

    @dispatch
    def minimum_clearance(self) -> pl.Series:
        """See [`GeoExprNameSpace.minimum_clearance`][polars_st.GeoExprNameSpace.minimum_clearance]."""  # noqa: E501
        ...

    @dispatch
    def x(self) -> pl.Series:
        """See [`GeoExprNameSpace.x`][polars_st.GeoExprNameSpace.x]."""
        ...

    @dispatch
    def y(self) -> pl.Series:
        """See [`GeoExprNameSpace.y`][polars_st.GeoExprNameSpace.y]."""
        ...

    @dispatch
    def z(self) -> pl.Series:
        """See [`GeoExprNameSpace.z`][polars_st.GeoExprNameSpace.z]."""
        ...

    @dispatch
    def m(self) -> pl.Series:
        """See [`GeoExprNameSpace.m`][polars_st.GeoExprNameSpace.m]."""
        ...

    @dispatch
    def count_coordinates(self) -> pl.Series:
        """See [`GeoExprNameSpace.count_coordinates`][polars_st.GeoExprNameSpace.count_coordinates]."""  # noqa: E501
        ...

    @dispatch
    def coordinates(self, output_dimension: CoordinateType | None = None) -> pl.Series:
        """See [`GeoExprNameSpace.coordinates`][polars_st.GeoExprNameSpace.coordinates]."""
        ...

    @dispatch
    def exterior_ring(self) -> GeoSeries:
        """See [`GeoExprNameSpace.exterior_ring`][polars_st.GeoExprNameSpace.exterior_ring]."""
        ...

    @dispatch
    def interior_rings(self) -> pl.Series:
        """See [`GeoExprNameSpace.interior_rings`][polars_st.GeoExprNameSpace.interior_rings]."""
        ...

    @dispatch
    def count_interior_rings(self) -> pl.Series:
        """See [`GeoExprNameSpace.count_interior_rings`][polars_st.GeoExprNameSpace.count_interior_rings]."""  # noqa: E501
        ...

    @dispatch
    def get_interior_ring(self, index: IntoIntegerExpr) -> GeoSeries:
        """See [`GeoExprNameSpace.get_interior_ring`][polars_st.GeoExprNameSpace.get_interior_ring]."""  # noqa: E501
        ...

    @dispatch
    def count_geometries(self) -> pl.Series:
        """See [`GeoExprNameSpace.count_geometries`][polars_st.GeoExprNameSpace.count_geometries]."""  # noqa: E501
        ...

    @dispatch
    def get_geometry(self, index: IntoIntegerExpr) -> GeoSeries:
        """See [`GeoExprNameSpace.get_geometry`][polars_st.GeoExprNameSpace.get_geometry]."""
        ...

    @dispatch
    def count_points(self) -> pl.Series:
        """See [`GeoExprNameSpace.count_points`][polars_st.GeoExprNameSpace.count_points]."""
        ...

    @dispatch
    def get_point(self, index: IntoIntegerExpr) -> GeoSeries:
        """See [`GeoExprNameSpace.get_point`][polars_st.GeoExprNameSpace.get_point]."""
        ...

    @dispatch
    def parts(self) -> pl.Series:
        """See [`GeoExprNameSpace.parts`][polars_st.GeoExprNameSpace.parts]."""
        ...

    @dispatch
    def precision(self) -> pl.Series:
        """See [`GeoExprNameSpace.precision`][polars_st.GeoExprNameSpace.precision]."""
        ...

    @dispatch
    def set_precision(
        self,
        grid_size: IntoNumericExpr,
        mode: Literal["valid_output", "no_topo", "keep_collapsed"] = "valid_output",
    ) -> GeoSeries:
        """See [`GeoExprNameSpace.set_precision`][polars_st.GeoExprNameSpace.set_precision]."""
        ...

    @dispatch
    def distance(self, other: IntoGeoExprColumn) -> pl.Series:
        """See [`GeoExprNameSpace.distance`][polars_st.GeoExprNameSpace.distance]."""
        ...

    @dispatch
    def hausdorff_distance(
        self,
        other: IntoGeoExprColumn,
        densify: float | None = None,
    ) -> pl.Series:
        """See [`GeoExprNameSpace.hausdorff_distance`][polars_st.GeoExprNameSpace.hausdorff_distance]."""  # noqa: E501
        ...

    @dispatch
    def frechet_distance(
        self,
        other: IntoGeoExprColumn,
        densify: float | None = None,
    ) -> pl.Series:
        """See [`GeoExprNameSpace.frechet_distance`][polars_st.GeoExprNameSpace.frechet_distance]."""  # noqa: E501
        ...

    # Projection operations

    @dispatch
    def srid(self) -> pl.Series:
        """See [`GeoExprNameSpace.srid`][polars_st.GeoExprNameSpace.srid]."""
        ...

    @dispatch
    def set_srid(self, srid: IntoIntegerExpr) -> GeoSeries:
        """See [`GeoExprNameSpace.set_srid`][polars_st.GeoExprNameSpace.set_srid]."""
        ...

    @dispatch
    def to_srid(self, srid: IntoIntegerExpr) -> GeoSeries:
        """See [`GeoExprNameSpace.to_srid`][polars_st.GeoExprNameSpace.to_srid]."""
        ...

    # Serialization

    @dispatch
    def to_wkt(
        self,
        rounding_precision: int | None = 6,
        trim: bool = True,
        output_dimension: Literal[2, 3, 4] = 3,
        old_3d: bool = False,
    ) -> pl.Series:
        """See [`GeoExprNameSpace.to_wkt`][polars_st.GeoExprNameSpace.to_wkt]."""
        ...

    @dispatch
    def to_ewkt(
        self,
        rounding_precision: int | None = 6,
        trim: bool = True,
        output_dimension: Literal[2, 3, 4] = 3,
        old_3d: bool = False,
    ) -> pl.Series:
        """See [`GeoExprNameSpace.to_ewkt`][polars_st.GeoExprNameSpace.to_ewkt]."""
        ...

    @dispatch
    def to_wkb(
        self,
        output_dimension: Literal[2, 3, 4] = 3,
        byte_order: Literal[0, 1] | None = None,
        include_srid: bool = False,
    ) -> pl.Series:
        """See [`GeoExprNameSpace.to_wkb`][polars_st.GeoExprNameSpace.to_wkb]."""
        ...

    @dispatch
    def to_geojson(self, indent: int | None = None) -> pl.Series:
        """See [`GeoExprNameSpace.to_geojson`][polars_st.GeoExprNameSpace.to_geojson]."""
        ...

    @dispatch
    def to_shapely(self) -> pl.Series:
        """See [`GeoExprNameSpace.to_shapely`][polars_st.GeoExprNameSpace.to_shapely]."""
        ...

    @dispatch
    def to_dict(self) -> pl.Series:
        """See [`GeoExprNameSpace.to_dict`][polars_st.GeoExprNameSpace.to_dict]."""
        ...

    @dispatch
    def cast(self, into: IntoExprColumn) -> pl.Series:
        """See [`GeoExprNameSpace.cast`][polars_st.GeoExprNameSpace.cast]."""
        ...

    @dispatch
    def multi(self) -> pl.Series:
        """See [`GeoExprNameSpace.multi`][polars_st.GeoExprNameSpace.multi]."""
        ...

    def to_geopandas(
        self,
        *,
        use_pyarrow_extension_array: bool = False,
        **kwargs: Any,
    ) -> gpd.GeoSeries:
        """Convert this pl.Series to a geopandas GeoSeries."""
        return (
            self._series
            .to_frame()
            .pipe(st)
            .to_geopandas(
                geometry_name=self._series.name,
                use_pyarrow_extension_array=use_pyarrow_extension_array,
                **kwargs,
            )
            .geometry
        )

    @property
    def __geo_interface__(self) -> dict:
        """Return a GeoJSON GeometryCollection [`dict`][] representation of the DataFrame."""
        return {
            "type": "GeometryCollection",
            "geometries": self.to_dict().to_list(),
        }

    #  Unary predicates

    @dispatch
    def has_z(self) -> pl.Series:
        """See [`GeoExprNameSpace.has_z`][polars_st.GeoExprNameSpace.has_z]."""
        ...

    @dispatch
    def has_m(self) -> pl.Series:
        """See [`GeoExprNameSpace.has_m`][polars_st.GeoExprNameSpace.has_m]."""
        ...

    @dispatch
    def is_ccw(self) -> pl.Series:
        """See [`GeoExprNameSpace.is_ccw`][polars_st.GeoExprNameSpace.is_ccw]."""
        ...

    @dispatch
    def is_closed(self) -> pl.Series:
        """See [`GeoExprNameSpace.is_closed`][polars_st.GeoExprNameSpace.is_closed]."""
        ...

    @dispatch
    def is_empty(self) -> pl.Series:
        """See [`GeoExprNameSpace.is_empty`][polars_st.GeoExprNameSpace.is_empty]."""
        ...

    @dispatch
    def is_ring(self) -> pl.Series:
        """See [`GeoExprNameSpace.is_ring`][polars_st.GeoExprNameSpace.is_ring]."""
        ...

    @dispatch
    def is_simple(self) -> pl.Series:
        """See [`GeoExprNameSpace.is_simple`][polars_st.GeoExprNameSpace.is_simple]."""
        ...

    @dispatch
    def is_valid(self) -> pl.Series:
        """See [`GeoExprNameSpace.is_valid`][polars_st.GeoExprNameSpace.is_valid]."""
        ...

    @dispatch
    def is_valid_reason(self) -> pl.Series:
        """See [`GeoExprNameSpace.is_valid_reason`][polars_st.GeoExprNameSpace.is_valid_reason]."""
        ...

    # Binary predicates

    @dispatch
    def crosses(self, other: IntoGeoExprColumn) -> pl.Series:
        """See [`GeoExprNameSpace.crosses`][polars_st.GeoExprNameSpace.crosses]."""
        ...

    @dispatch
    def contains(self, other: IntoGeoExprColumn) -> pl.Series:
        """See [`GeoExprNameSpace.contains`][polars_st.GeoExprNameSpace.contains]."""
        ...

    @dispatch
    def contains_properly(self, other: IntoGeoExprColumn) -> pl.Series:
        """See [`GeoExprNameSpace.contains_properly`][polars_st.GeoExprNameSpace.contains_properly]."""  # noqa: E501
        ...

    @dispatch
    def covered_by(self, other: IntoGeoExprColumn) -> pl.Series:
        """See [`GeoExprNameSpace.covered_by`][polars_st.GeoExprNameSpace.covered_by]."""
        ...

    @dispatch
    def covers(self, other: IntoGeoExprColumn) -> pl.Series:
        """See [`GeoExprNameSpace.covers`][polars_st.GeoExprNameSpace.covers]."""
        ...

    @dispatch
    def disjoint(self, other: IntoGeoExprColumn) -> pl.Series:
        """See [`GeoExprNameSpace.disjoint`][polars_st.GeoExprNameSpace.disjoint]."""
        ...

    @dispatch
    def dwithin(self, other: IntoGeoExprColumn, distance: IntoNumericExpr) -> pl.Series:
        """See [`GeoExprNameSpace.dwithin`][polars_st.GeoExprNameSpace.dwithin]."""
        ...

    @dispatch
    def intersects(self, other: IntoGeoExprColumn) -> pl.Series:
        """See [`GeoExprNameSpace.intersects`][polars_st.GeoExprNameSpace.intersects]."""
        ...

    @dispatch
    def overlaps(self, other: IntoGeoExprColumn) -> pl.Series:
        """See [`GeoExprNameSpace.overlaps`][polars_st.GeoExprNameSpace.overlaps]."""
        ...

    @dispatch
    def touches(self, other: IntoGeoExprColumn) -> pl.Series:
        """See [`GeoExprNameSpace.touches`][polars_st.GeoExprNameSpace.touches]."""
        ...

    @dispatch
    def within(self, other: IntoGeoExprColumn) -> pl.Series:
        """See [`GeoExprNameSpace.within`][polars_st.GeoExprNameSpace.within]."""
        ...

    @dispatch
    def equals(self, other: IntoGeoExprColumn) -> pl.Series:
        """See [`GeoExprNameSpace.equals`][polars_st.GeoExprNameSpace.equals]."""
        ...

    @dispatch
    def equals_exact(
        self,
        other: IntoGeoExprColumn,
        tolerance: float = 0.0,
    ) -> pl.Series:
        """See [`GeoExprNameSpace.equals_exact`][polars_st.GeoExprNameSpace.equals_exact]."""
        ...

    @dispatch
    def equals_identical(self, other: IntoGeoExprColumn) -> pl.Series:
        """See [`GeoExprNameSpace.equals_identical`][polars_st.GeoExprNameSpace.equals_identical]."""  # noqa: E501
        ...

    @dispatch
    def relate(self, other: IntoGeoExprColumn) -> pl.Series:
        """See [`GeoExprNameSpace.relate`][polars_st.GeoExprNameSpace.relate]."""
        ...

    @dispatch
    def relate_pattern(
        self,
        other: IntoGeoExprColumn,
        pattern: str,
    ) -> pl.Series:
        """See [`GeoExprNameSpace.relate_pattern`][polars_st.GeoExprNameSpace.relate_pattern]."""
        ...

    # Set operations

    @dispatch
    def union(
        self,
        other: IntoGeoExprColumn,
        grid_size: float | None = None,
    ) -> GeoSeries:
        """See [`GeoExprNameSpace.union`][polars_st.GeoExprNameSpace.union]."""
        ...

    @dispatch
    def unary_union(self, grid_size: float | None = None) -> GeoSeries:
        """See [`GeoExprNameSpace.unary_union`][polars_st.GeoExprNameSpace.unary_union]."""
        ...

    @dispatch
    def coverage_union(self) -> GeoSeries:
        """See [`GeoExprNameSpace.coverage_union`][polars_st.GeoExprNameSpace.coverage_union]."""
        ...

    @dispatch
    def intersection(
        self,
        other: IntoGeoExprColumn,
        grid_size: float | None = None,
    ) -> GeoSeries:
        """See [`GeoExprNameSpace.intersection`][polars_st.GeoExprNameSpace.intersection]."""
        ...

    @dispatch
    def difference(
        self,
        other: IntoGeoExprColumn,
        grid_size: float | None = None,
    ) -> GeoSeries:
        """See [`GeoExprNameSpace.difference`][polars_st.GeoExprNameSpace.difference]."""
        ...

    @dispatch
    def symmetric_difference(
        self,
        other: IntoGeoExprColumn,
        grid_size: float | None = None,
    ) -> GeoSeries:
        """See [`GeoExprNameSpace.symmetric_difference`][polars_st.GeoExprNameSpace.symmetric_difference]."""  # noqa: E501
        ...

    # Constructive operations

    @dispatch
    def boundary(self) -> GeoSeries:
        """See [`GeoExprNameSpace.boundary`][polars_st.GeoExprNameSpace.boundary]."""
        ...

    @dispatch
    def buffer(
        self,
        distance: IntoNumericExpr,
        quad_segs: int = 8,
        cap_style: Literal["round", "square", "flat"] = "round",
        join_style: Literal["round", "mitre", "bevel"] = "round",
        mitre_limit: float = 5.0,
        single_sided: bool = False,
    ) -> GeoSeries:
        """See [`GeoExprNameSpace.buffer`][polars_st.GeoExprNameSpace.buffer]."""
        ...

    @dispatch
    def offset_curve(
        self,
        distance: IntoNumericExpr,
        quad_segs: int = 8,
        join_style: Literal["round", "mitre", "bevel"] = "round",
        mitre_limit: float = 5.0,
    ) -> GeoSeries:
        """See [`GeoExprNameSpace.offset_curve`][polars_st.GeoExprNameSpace.offset_curve]."""
        ...

    @dispatch
    def centroid(self) -> GeoSeries:
        """See [`GeoExprNameSpace.centroid`][polars_st.GeoExprNameSpace.centroid]."""
        ...

    @dispatch
    def center(self) -> GeoSeries:
        """See [`GeoExprNameSpace.center`][polars_st.GeoExprNameSpace.center]."""
        ...

    @dispatch
    def clip_by_rect(self, bounds: IntoExprColumn) -> GeoSeries:
        """See [`GeoExprNameSpace.clip_by_rect`][polars_st.GeoExprNameSpace.clip_by_rect]."""
        ...

    @dispatch
    def convex_hull(self) -> GeoSeries:
        """See [`GeoExprNameSpace.convex_hull`][polars_st.GeoExprNameSpace.convex_hull]."""
        ...

    @dispatch
    def concave_hull(self, ratio: float = 0.0, allow_holes: bool = False) -> GeoSeries:
        """See [`GeoExprNameSpace.concave_hull`][polars_st.GeoExprNameSpace.concave_hull]."""
        ...

    @dispatch
    def segmentize(self, max_segment_length: IntoNumericExpr) -> GeoSeries:
        """See [`GeoExprNameSpace.segmentize`][polars_st.GeoExprNameSpace.segmentize]."""
        ...

    @dispatch
    def envelope(self) -> GeoSeries:
        """See [`GeoExprNameSpace.envelope`][polars_st.GeoExprNameSpace.envelope]."""
        ...

    @dispatch
    def extract_unique_points(self) -> GeoSeries:
        """See [`GeoExprNameSpace.extract_unique_points`][polars_st.GeoExprNameSpace.extract_unique_points]."""  # noqa: E501
        ...

    @dispatch
    def build_area(self) -> GeoSeries:
        """See [`GeoExprNameSpace.build_area`][polars_st.GeoExprNameSpace.build_area]."""
        ...

    @dispatch
    def make_valid(
        self,
        method: Literal["linework", "structure"] = "linework",
        keep_collapsed: bool = True,
    ) -> GeoSeries:
        """See [`GeoExprNameSpace.make_valid`][polars_st.GeoExprNameSpace.make_valid]."""
        ...

    @dispatch
    def normalize(self) -> GeoSeries:
        """See [`GeoExprNameSpace.normalize`][polars_st.GeoExprNameSpace.normalize]."""
        ...

    @dispatch
    def node(self) -> GeoSeries:
        """See [`GeoExprNameSpace.node`][polars_st.GeoExprNameSpace.node]."""
        ...

    @dispatch
    def point_on_surface(self) -> GeoSeries:
        """See [`GeoExprNameSpace.point_on_surface`][polars_st.GeoExprNameSpace.point_on_surface]."""  # noqa: E501
        ...

    @dispatch
    def remove_repeated_points(self, tolerance: IntoNumericExpr = 0.0) -> GeoSeries:
        """See [`GeoExprNameSpace.remove_repeated_points`][polars_st.GeoExprNameSpace.remove_repeated_points]."""  # noqa: E501
        ...

    @dispatch
    def reverse(self) -> GeoSeries:
        """See [`GeoExprNameSpace.reverse`][polars_st.GeoExprNameSpace.reverse]."""
        ...

    @dispatch
    def simplify(
        self,
        tolerance: IntoNumericExpr,
        preserve_topology: bool = True,
    ) -> GeoSeries:
        """See [`GeoExprNameSpace.simplify`][polars_st.GeoExprNameSpace.simplify]."""
        ...

    @dispatch
    def force_2d(self) -> GeoSeries:
        """See [`GeoExprNameSpace.force_2d`][polars_st.GeoExprNameSpace.force_2d]."""
        ...

    @dispatch
    def force_3d(self, z: IntoNumericExpr = 0.0) -> GeoSeries:
        """See [`GeoExprNameSpace.force_3d`][polars_st.GeoExprNameSpace.force_3d]."""
        ...

    @dispatch
    def flip_coordinates(self) -> GeoSeries:
        """See [`GeoExprNameSpace.flip_coordinates`][polars_st.GeoExprNameSpace.flip_coordinates]."""  # noqa: E501
        ...

    @dispatch
    def minimum_rotated_rectangle(self) -> GeoSeries:
        """See [`GeoExprNameSpace.minimum_rotated_rectangle`][polars_st.GeoExprNameSpace.minimum_rotated_rectangle]."""  # noqa: E501
        ...

    @dispatch
    def maximum_inscribed_circle(self, tolerance: IntoNumericExpr) -> GeoSeries:
        """See [`GeoExprNameSpace.maximum_inscribed_circle`][polars_st.GeoExprNameSpace.maximum_inscribed_circle]."""  # noqa: E501
        ...

    @dispatch
    def snap(
        self,
        other: IntoGeoExprColumn,
        tolerance: IntoNumericExpr,
    ) -> GeoSeries:
        """See [`GeoExprNameSpace.snap`][polars_st.GeoExprNameSpace.snap]."""
        ...

    @dispatch
    def shortest_line(self, other: IntoGeoExprColumn) -> GeoSeries:
        """See [`GeoExprNameSpace.shortest_line`][polars_st.GeoExprNameSpace.shortest_line]."""
        ...

    # Affine transforms

    @dispatch
    def affine_transform(self, matrix: IntoExprColumn | Sequence[float]) -> GeoSeries:
        """See [`GeoExprNameSpace.affine_transform`][polars_st.GeoExprNameSpace.affine_transform]."""  # noqa: E501
        ...

    @dispatch
    def translate(
        self,
        x: IntoNumericExpr = 0.0,
        y: IntoNumericExpr = 0.0,
        z: IntoNumericExpr = 0.0,
    ) -> GeoSeries:
        """See [`GeoExprNameSpace.translate`][polars_st.GeoExprNameSpace.translate]."""
        ...

    @dispatch
    def rotate(
        self,
        angle: IntoNumericExpr,
        origin: Literal["center", "centroid"] | Sequence[float] = "center",
    ) -> GeoSeries:
        """See [`GeoExprNameSpace.rotate`][polars_st.GeoExprNameSpace.rotate]."""
        ...

    @dispatch
    def scale(
        self,
        x: IntoNumericExpr = 1.0,
        y: IntoNumericExpr = 1.0,
        z: IntoNumericExpr = 1.0,
        origin: Literal["center", "centroid"] | Sequence[float] = "center",
    ) -> GeoSeries:
        """See [`GeoExprNameSpace.scale`][polars_st.GeoExprNameSpace.scale]."""
        ...

    @dispatch
    def skew(
        self,
        x: IntoNumericExpr = 0.0,
        y: IntoNumericExpr = 0.0,
        z: IntoNumericExpr = 0.0,
        origin: Literal["center", "centroid"] | Sequence[float] = "center",
    ) -> GeoSeries:
        """See [`GeoExprNameSpace.skew`][polars_st.GeoExprNameSpace.skew]."""
        ...

    # LineString operations

    @dispatch
    def interpolate(
        self,
        distance: IntoNumericExpr,
        normalized: bool = False,
    ) -> GeoSeries:
        """See [`GeoExprNameSpace.interpolate`][polars_st.GeoExprNameSpace.interpolate]."""
        ...

    @dispatch
    def project(
        self,
        other: IntoGeoExprColumn,
        normalized: bool = False,
    ) -> pl.Series:
        """See [`GeoExprNameSpace.project`][polars_st.GeoExprNameSpace.project]."""
        ...

    @dispatch
    def substring(
        self,
        start: IntoNumericExpr,
        end: IntoNumericExpr,
    ) -> GeoSeries:
        """See [`GeoExprNameSpace.substring`][polars_st.GeoExprNameSpace.substring]."""
        ...

    @dispatch
    def line_merge(self, directed: bool = False) -> GeoSeries:
        """See [`GeoExprNameSpace.line_merge`][polars_st.GeoExprNameSpace.line_merge]."""
        ...

    @dispatch
    def shared_paths(
        self,
        other: IntoGeoExprColumn,
    ) -> GeoSeries:
        """See [`GeoExprNameSpace.shared_paths`][polars_st.GeoExprNameSpace.shared_paths]."""
        ...

    # Aggregations

    @dispatch
    def total_bounds(self) -> pl.Series:
        """See [`GeoExprNameSpace.total_bounds`][polars_st.GeoExprNameSpace.total_bounds]."""
        ...

    @dispatch
    def collect(self, into: GeometryType | None = None) -> GeoSeries:
        """See [`GeoExprNameSpace.collect`][polars_st.GeoExprNameSpace.collect]."""
        ...

    @dispatch
    def union_all(self, grid_size: float | None = None) -> GeoSeries:
        """See [`GeoExprNameSpace.union_all`][polars_st.GeoExprNameSpace.union_all]."""
        ...

    @dispatch
    def coverage_union_all(self) -> GeoSeries:
        """See [`GeoExprNameSpace.coverage_union_all`][polars_st.GeoExprNameSpace.coverage_union_all]."""  # noqa: E501
        ...

    @dispatch
    def intersection_all(self, grid_size: float | None = None) -> GeoSeries:
        """See [`GeoExprNameSpace.intersection_all`][polars_st.GeoExprNameSpace.intersection_all]."""  # noqa: E501
        ...

    @dispatch
    def difference_all(self, grid_size: float | None = None) -> GeoSeries:
        """See [`GeoExprNameSpace.difference_all`][polars_st.GeoExprNameSpace.difference_all]."""
        ...

    @dispatch
    def symmetric_difference_all(self, grid_size: float | None = None) -> GeoSeries:
        """See [`GeoExprNameSpace.symmetric_difference_all`][polars_st.GeoExprNameSpace.symmetric_difference_all]."""  # noqa: E501
        ...

    @dispatch
    def polygonize(self) -> GeoSeries:
        """See [`GeoExprNameSpace.polygonize`][polars_st.GeoExprNameSpace.polygonize]."""
        ...

    @dispatch
    def voronoi_polygons(
        self,
        tolerance: float = 0.0,
        extend_to: bytes | None = None,
        only_edges: bool = False,
    ) -> GeoSeries:
        """See [`GeoExprNameSpace.voronoi_polygons`][polars_st.GeoExprNameSpace.voronoi_polygons]."""  # noqa: E501
        ...

    @dispatch
    def delaunay_triangles(
        self,
        tolerance: float = 0.0,
        only_edges: bool = False,
    ) -> GeoSeries:
        """See [`GeoExprNameSpace.delaunay_triangles`][polars_st.GeoExprNameSpace.delaunay_triangles]."""  # noqa: E501
        ...

    def plot(self, **kwargs: Unpack[MarkConfigKwds]) -> alt.Chart:
        """Draw map plot.

        Polars does not implement plotting logic itself but instead defers to
        [`Altair`](https://altair-viz.github.io/).

        `df.st.plot(**kwargs)` is shorthand for
        `alt.Chart(df).mark_geoshape(**kwargs).interactive()`. Please read Altair
        [GeoShape](https://altair-viz.github.io/user_guide/marks/geoshape.html) documentation
        for available options.
        """
        return self._series.to_frame().pipe(st).plot(**kwargs)

    def explore(
        self,
        *,
        scatterplot_kwargs: ScatterplotLayerKwargs | None = None,
        path_kwargs: PathLayerKwargs | None = None,
        polygon_kwargs: PolygonLayerKwargs | None = None,
        map_kwargs: MapKwargs | None = None,
    ) -> Map:
        return (
            self._series
            .to_frame()
            .pipe(st)
            .explore(
                self._series.name,
                scatterplot_kwargs=scatterplot_kwargs,
                path_kwargs=path_kwargs,
                polygon_kwargs=polygon_kwargs,
                map_kwargs=map_kwargs,
            )
        )



import pprint

import polars as pl
import pytest

import polars_st as st


@pytest.fixture(autouse=True)
def setup(doctest_namespace: dict) -> None:
    doctest_namespace["pl"] = pl
    doctest_namespace["st"] = st
    doctest_namespace["pprint"] = pprint



from __future__ import annotations

from typing import TYPE_CHECKING, cast, overload

import polars as pl
from polars._dependencies import pandas as pd
from pyogrio import read_arrow

from polars_st.casting import st
from polars_st.selectors import geom
from polars_st.utils.srid import get_crs_srid_or_warn

if TYPE_CHECKING:
    from collections.abc import Sequence
    from pathlib import Path

    import geopandas as gpd
    from polars._typing import SchemaDict

    from polars_st.geodataframe import GeoDataFrame
    from polars_st.geoseries import GeoSeries


__all__ = [
    "from_geopandas",
    "read_file",
]


def read_file(
    path_or_buffer: Path | str | bytes,
    /,
    layer: int | str | None = None,
    encoding: str | None = None,
    columns: Sequence[str] | None = None,
    read_geometry: bool = True,
    force_2d: bool = False,
    skip_features: int = 0,
    max_features: int | None = None,
    where: str | None = None,
    bbox: tuple[float, float, float, float] | None = None,
    fids: Sequence[int] | None = None,
    sql: str | None = None,
    sql_dialect: str | None = None,
    return_fids: bool = False,
) -> GeoDataFrame:
    """Read OGR data source into a GeoDataFrame.

    IMPORTANT: non-linear geometry types (e.g., MultiSurface) are converted
    to their linear approximations.

    Args:
        path_or_buffer: A dataset path or URI, or raw buffer.
        layer:
            If an integer is provided, it corresponds to the index of the layer
            with the data source.  If a string is provided, it must match the name
            of the layer in the data source.  Defaults to first layer in data source.
        encoding:
            If present, will be used as the encoding for reading string values from
            the data source, unless encoding can be inferred directly from the data
            source.
        columns:
            List of column names to import from the data source.  Column names must
            exactly match the names in the data source, and will be returned in
            the order they occur in the data source.  To avoid reading any columns,
            pass an empty list-like.
        read_geometry:
            If True, will read geometry into WKB. If False, geometry will be None.
            Defaults to `True`.
        force_2d:
            If the geometry has Z values, setting this to True will cause those to
            be ignored and 2D geometries to be returned. Defaults to `False`.
        skip_features:
            Number of features to skip from the beginning of the file before returning
            features.  Must be less than the total number of features in the file.
        max_features:
            Number of features to read from the file.  Must be less than the total
            number of features in the file minus skip_features (if used).
        where:
            Where clause to filter features in layer by attribute values.  Uses a
            restricted form of SQL WHERE clause, defined [here](http://ogdi.sourceforge.net/prop/6.2.CapabilitiesMetadata.html).
            For examples:

            - `"ISO_A3 = 'CAN'"`
            - `"POP_EST > 10000000 AND POP_EST < 100000000"`
        bbox:
            If present, will be used to filter records whose geometry intersects this
            box.  This must be in the same CRS as the dataset.  If GEOS is present
            and used by GDAL, only geometries that intersect this bbox will be
            returned; if GEOS is not available or not used by GDAL, all geometries
            with bounding boxes that intersect this bbox will be returned.
        fids:
            Array of integer feature id (FID) values to select. Cannot be combined
            with other keywords to select a subset (`skip_features`, `max_features`,
            `where` or `bbox`). Note that the starting index is driver and file
            specific (e.g. typically 0 for Shapefile and 1 for GeoPackage, but can
            still depend on the specific file). The performance of reading a large
            number of features usings FIDs is also driver specific.
        sql:
            The SQL statement to execute. Look at the sql_dialect parameter for more
            information on the syntax to use for the query. When combined with other
            keywords like `columns`, `skip_features`, `max_features`,
            `where`, `bbox`, or `mask`, those are applied after the SQL query.
            Be aware that this can have an impact on performance, (e.g. filtering
            with the `bbox` or `mask` keywords may not use spatial indexes).
            Cannot be combined with the `layer` or `fids` keywords.
        sql_dialect:
            The SQL dialect the SQL statement is written in. Possible values:

            - `None`: if the data source natively supports SQL, its specific SQL dialect
                will be used by default (eg. SQLite and Geopackage: `SQLITE`_, PostgreSQL).
                If the data source doesn't natively support SQL, the `OGRSQL`_ dialect is
                the default.
            - `OGRSQL`: can be used on any data source. Performance can suffer
                when used on data sources with native support for SQL.
            - `SQLITE`: can be used on any data source. All spatialite_
                functions can be used. Performance can suffer on data sources with
                native support for SQL, except for Geopackage and SQLite as this is
                their native SQL dialect.
        return_fids:
            If True, will return the FIDs of the feature that were read.
    """
    metadata, table = read_arrow(
        path_or_buffer,
        layer=layer,
        encoding=encoding,
        columns=columns,
        read_geometry=read_geometry,
        force_2d=force_2d,
        skip_features=skip_features,
        max_features=max_features,
        where=where,
        bbox=bbox,
        fids=fids,
        sql=sql,
        sql_dialect=sql_dialect,
        return_fids=return_fids,
    )

    import pyarrow as pa

    geometry_name = metadata["geometry_name"] or "wkb_geometry"
    if geometry_name in table.column_names:
        geometry_index = table.schema.get_field_index(geometry_name)
        table = table.cast(table.schema.set(geometry_index, pa.field(geometry_name, pa.binary())))
        res = cast("pl.DataFrame", pl.from_arrow(table))
        if (crs := metadata["crs"]) and (srid := get_crs_srid_or_warn(crs)):
            res = res.with_columns(geom(geometry_name).st.set_srid(srid))
        if not metadata["geometry_name"]:
            res = res.rename({"wkb_geometry": "geometry"})
    else:
        res = cast("pl.DataFrame", pl.from_arrow(table))
    return st(res)._df  # noqa: SLF001


@overload
def from_geopandas(
    data: gpd.GeoDataFrame,
    *,
    schema_overrides: SchemaDict | None = None,
    rechunk: bool = True,
    nan_to_null: bool = True,
    include_index: bool = False,
) -> GeoDataFrame: ...


@overload
def from_geopandas(
    data: gpd.GeoSeries,
    *,
    schema_overrides: SchemaDict | None = None,
    rechunk: bool = True,
    nan_to_null: bool = True,
    include_index: bool = False,
) -> GeoSeries: ...


def from_geopandas(
    data: gpd.GeoDataFrame | gpd.GeoSeries,
    *,
    schema_overrides: SchemaDict | None = None,
    rechunk: bool = True,
    nan_to_null: bool = True,
    include_index: bool = False,
) -> GeoDataFrame | GeoSeries:
    """Create DataFrame or Series from Geopandas GeoDataFrame or GeoSeries.

    Examples:
        >>> import shapely
        >>> import geopandas as gpd
        >>> pd_gdf = gpd.GeoDataFrame({
        ...     "geometry": [shapely.Point(0, 0), shapely.Point(1, 2)]
        ... }, crs="EPSG:4326")
        >>> gdf = st.from_geopandas(pd_gdf)
    """
    res = pl.from_pandas(
        data.to_wkb(),
        schema_overrides=schema_overrides,
        rechunk=rechunk,
        nan_to_null=nan_to_null,
        include_index=include_index,
    )

    if isinstance(data, pd.Series):
        res = pl.from_pandas(
            data.to_wkb(),
            schema_overrides=schema_overrides,
            rechunk=rechunk,
            nan_to_null=nan_to_null,
            include_index=include_index,
        )
        res = cast("pl.Series", res)
        if (crs := data.crs) and (srid := get_crs_srid_or_warn(str(crs))):
            res = st(res).set_srid(srid)
        return st(res)._series  # noqa: SLF001

    res = cast("pl.DataFrame", res).with_columns(
        geom(str(col)).st.set_srid(srid)
        for col in data.dtypes.index[data.dtypes == "geometry"]
        if (crs := data[col].crs) and (srid := get_crs_srid_or_warn(str(crs)))
    )
    return st(res)._df  # noqa: SLF001



from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeAlias, Union

if TYPE_CHECKING:
    from collections.abc import Sequence

    import numpy as np
    import pandas as pd
    import polars as pl
    import pyarrow as pa
    from polars._typing import (
        ArrowArrayExportable,
        ArrowStreamExportable,
    )

    ArrayLike = Union[  # noqa: UP007
        Sequence[Any],
        pl.Series,
        pa.Array,
        pa.ChunkedArray,
        np.ndarray[Any, Any],
        pd.Series,
        pd.DatetimeIndex,
        ArrowArrayExportable,
        ArrowStreamExportable,
    ]

    IntoExprColumn: TypeAlias = pl.Expr | pl.Series | str
    IntoGeoExprColumn: TypeAlias = IntoExprColumn
    IntoIntegerExpr: TypeAlias = IntoExprColumn | int
    IntoNumericExpr: TypeAlias = IntoExprColumn | int | float



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Literal, cast

import polars as pl
from polars import Expr, LazyFrame
from polars._utils.parse import parse_into_expression
from polars.api import register_lazyframe_namespace
from polars.datatypes import N_INFER_DEFAULT
from polars.plugins import register_plugin_function

from polars_st.geodataframe import GeoDataFrame

if TYPE_CHECKING:
    from polars._typing import (
        FrameInitTypes,
        JoinStrategy,
        JoinValidation,
        MaintainOrderJoin,
        Orientation,
        SchemaDefinition,
        SchemaDict,
    )


__all__ = [
    "GeoLazyFrame",
    "GeoLazyFrameNameSpace",
]


class GeoLazyFrame(LazyFrame):
    @property
    def st(self) -> GeoLazyFrameNameSpace: ...

    def __new__(  # noqa: PYI034
        cls,
        data: FrameInitTypes | None = None,
        schema: SchemaDefinition | None = None,
        *,
        schema_overrides: SchemaDict | None = None,
        strict: bool = True,
        orient: Orientation | None = None,
        infer_schema_length: int | None = N_INFER_DEFAULT,
        nan_to_null: bool = False,
    ) -> GeoLazyFrame:
        ldf = GeoDataFrame(
            data=data,
            schema=schema,
            schema_overrides=schema_overrides,
            strict=strict,
            orient=orient,
            infer_schema_length=infer_schema_length,
            nan_to_null=nan_to_null,
        ).lazy()
        return cast("GeoLazyFrame", ldf)


@register_lazyframe_namespace("st")
class GeoLazyFrameNameSpace:
    def __init__(self, lf: LazyFrame) -> None:
        self._lf = lf

    def sjoin(
        self,
        other: LazyFrame,
        on: str | Expr = "geometry",
        how: JoinStrategy = "inner",
        predicate: Literal[
            "intersects_bbox",
            "intersects",
            "within",
            "dwithin",
            "contains",
            "overlaps",
            "crosses",
            "touches",
            "covers",
            "covered_by",
            "contains_properly",
        ] = "intersects",
        distance: float | None = None,
        *,
        left_on: str | Expr | None = None,
        right_on: str | Expr | None = None,
        suffix: str = "_right",
        validate: JoinValidation = "m:m",
        nulls_equal: bool = False,
        coalesce: bool | None = None,
        maintain_order: MaintainOrderJoin | None = None,
        allow_parallel: bool = True,
        force_parallel: bool = False,
    ) -> LazyFrame:
        """Perform a spatial join operation with another LazyFrame."""
        if not isinstance(other, LazyFrame):
            msg = f"expected `other` join table to be a LazyFrame, not a {type(other).__name__!r}"
            raise TypeError(msg)

        if how == "cross":
            msg = """Use of `how="cross" not supported on sjoin.`"""
            raise ValueError(msg)

        left_expr = left_on or on
        right_expr = right_on or on

        if (
            parse_into_expression(left_expr).meta_has_multiple_outputs()
            or parse_into_expression(right_expr).meta_has_multiple_outputs()
        ):
            msg = "spatial join expressions should not return multiple output"
            raise ValueError(msg)

        sjoin_index = (
            pl
            .concat(
                [
                    self._lf.select(_sjoin_geom_left=left_expr),
                    other.select(_sjoin_geom_right=right_expr),
                ],
                how="horizontal",
            )
            .select(
                register_plugin_function(
                    plugin_path=Path(__file__).parent,
                    function_name="sjoin",
                    args=["_sjoin_geom_left", "_sjoin_geom_right"],
                    kwargs={"predicate": {"type": predicate, "param": distance}},
                    is_elementwise=True,
                ),
            )
            .select(
                _sjoin_index_left=pl.nth(0).struct[0],
                _sjoin_index_right=pl.nth(0).struct[1],
            )
        )

        return (
            sjoin_index
            .join(
                self._lf.with_row_index("_sjoin_index_left"),
                on="_sjoin_index_left",
                how="full",
                suffix=suffix,
                coalesce=coalesce,
                maintain_order=maintain_order,
                allow_parallel=allow_parallel,
                force_parallel=force_parallel,
            )
            .join(
                other.with_row_index("_sjoin_index_right"),
                on="_sjoin_index_right",
                how=how,
                validate=validate,
                nulls_equal=nulls_equal,
                coalesce=coalesce,
                maintain_order=maintain_order,
                allow_parallel=allow_parallel,
                force_parallel=force_parallel,
            )
            .drop(
                "_sjoin_index_left",
                "_sjoin_index_right",
                "_sjoin_index_left_right",
            )
        )



from __future__ import annotations

from typing import Literal, TypeAlias, get_args

import polars as pl

__all__ = [
    "CoordinateType",
    "DimensionType",
    "GeometryType",
    "PolarsCoordinateType",
    "PolarsDimensionType",
    "PolarsGeometryType",
]

GeometryType: TypeAlias = Literal[
    "Unknown",
    "Point",
    "LineString",
    "Polygon",
    "MultiPoint",
    "MultiLineString",
    "MultiPolygon",
    "GeometryCollection",
    "CircularString",
    "CompoundCurve",
    "CurvePolygon",
    "MultiCurve",
    "MultiSurface",
    "Curve",
    "Surface",
    "PolyhedralSurface",
    "Tin",
    "Triangle",
]

DimensionType: TypeAlias = Literal[
    "Point",
    "Curve",
    "Surface",
]

CoordinateType: TypeAlias = Literal[
    "XY",
    "XYZ",
    "XYZM",
    "XYM",
]

PolarsGeometryType = pl.Enum(get_args(GeometryType))
PolarsDimensionType = pl.Enum(get_args(DimensionType))
PolarsCoordinateType = pl.Enum(get_args(CoordinateType))



from __future__ import annotations

import logging

from polars_st._lib import get_crs_authority

logger = logging.getLogger(__name__)


def get_crs_srid_or_warn(crs: str) -> int | None:
    authority = get_crs_authority(crs)
    if authority is None:
        msg = f"Couldn't infer authority from {crs}.\n The geometries SRID will be set to 0."
        logger.warning(msg)
        return None

    name, code = authority
    if not code.isdigit():
        msg = f'Couldn\'t convert authority "{name}:{code}" to an integer srid.'
        logger.warning(msg)
        return None

    return int(code, base=10)



from __future__ import annotations

from functools import wraps
from inspect import signature
from pathlib import Path
from typing import TYPE_CHECKING, Literal, ParamSpec, cast

import polars as pl
from polars._utils.parse import parse_into_expression
from polars._utils.wrap import wrap_expr
from polars.api import register_expr_namespace
from polars.plugins import register_plugin_function

from polars_st import _lib
from polars_st.geometry import (
    CoordinateType,
    GeometryType,
    PolarsCoordinateType,
    PolarsDimensionType,
    PolarsGeometryType,
)
from polars_st.utils.internal import is_empty_method

if TYPE_CHECKING:
    from collections.abc import Sequence

    from polars_st.typing import (
        IntoExprColumn,
        IntoGeoExprColumn,
        IntoIntegerExpr,
        IntoNumericExpr,
    )

    P = ParamSpec("P")

__all__ = [
    "GeoExpr",
    "GeoExprNameSpace",
]


def register_plugin(is_aggregation: bool = False, cast: pl.DataType | None = None):  # noqa: ANN202
    def decorator(func):  # noqa: ANN001, ANN202
        assert is_empty_method(func)  # noqa: S101

        func_name = func.__name__
        sig = signature(func)
        params = sig.parameters
        expr_args = dict.fromkeys(k for k, v in params.items() if "Expr" in str(v.annotation))
        other_args = dict.fromkeys(k for k, v in params.items() if k not in expr_args)

        @wraps(func)
        def wrapper(self: GeoExprNameSpace, *args: P.args, **kwargs: P.kwargs):  # noqa: ANN202
            bound = sig.bind(self._expr, *args, **kwargs)
            bound.apply_defaults()
            result = register_plugin_function(
                plugin_path=Path(__file__).parent,
                function_name=func_name,
                args=[self._expr, *[bound.arguments[k] for k in expr_args]],
                kwargs={k: bound.arguments[k] for k in other_args},
                is_elementwise=not is_aggregation,
                returns_scalar=is_aggregation,
            )
            return result.cast(cast) if cast is not None else result

        return wrapper

    return decorator


class GeoExpr(pl.Expr):
    """`GeoExpr` is used as an alias for [`polars.Expr`](https://docs.pola.rs/api/python/stable/reference/expressions/index.html) with type annotations added for the `st` namespace."""  # noqa: E501

    @property
    def st(self) -> GeoExprNameSpace:
        return GeoExprNameSpace(self)

    def __new__(cls) -> GeoExpr:  # noqa: PYI034
        return cast("GeoExpr", pl.Expr())


@register_expr_namespace("st")
class GeoExprNameSpace:
    def __init__(self, expr: pl.Expr) -> None:
        self._expr = cast("GeoExpr", expr)

    @register_plugin(cast=PolarsGeometryType)
    def geometry_type(self) -> pl.Expr:
        """Return the type of each geometry as [`PolarsGeometryType`][polars_st.PolarsGeometryType].

        Examples:
            >>> gdf = st.GeoDataFrame([
            ...     "POINT(0 0)",
            ...     "LINESTRING(0 0, 1 2)",
            ...     "POLYGON((0 0, 1 1, 1 0, 0 0))"
            ... ])
            >>> gdf.select(st.geom().st.geometry_type())
            shape: (3, 1)
            ┌────────────┐
            │ geometry   │
            │ ---        │
            │ enum       │
            ╞════════════╡
            │ Point      │
            │ LineString │
            │ Polygon    │
            └────────────┘
        """
        ...

    @register_plugin(cast=PolarsDimensionType)
    def dimension(self) -> pl.Expr:
        """Return the inherent dimensionality of each geometry.

        The inherent dimension is 0 for points, 1 for linestrings and linearrings,
            and 2 for polygons. For geometrycollections it is the max of the containing
            elements.

        Examples:
            >>> gdf = st.GeoDataFrame([
            ...     "POINT(0 0)",
            ...     "LINESTRING(0 0, 1 2)",
            ...     "POLYGON((0 0, 1 1, 1 0, 0 0))"
            ... ])
            >>> gdf.select(st.geom().st.dimension())
            shape: (3, 1)
            ┌──────────┐
            │ geometry │
            │ ---      │
            │ enum     │
            ╞══════════╡
            │ Point    │
            │ Curve    │
            │ Surface  │
            └──────────┘
        """
        ...

    @register_plugin()
    def coordinate_dimension(self) -> pl.Expr:
        """Return the coordinate dimension (2, 3 or 4) of each geometry."""
        ...

    @register_plugin(cast=PolarsCoordinateType)
    def coordinate_type(self) -> pl.Expr:
        """Return the coordinate type of each geometry as [`PolarsCoordinateType`][polars_st.PolarsCoordinateType].

        Examples:
            >>> gdf = st.GeoDataFrame([
            ...     "POINT(1 2)",
            ...     "POINT(1 2 3)",
            ...     "POINT M(1 2 3)",
            ...     "POINT(1 2 3 4)",
            ... ])
            >>> gdf.select(st.geom().st.coordinate_type())
            shape: (4, 1)
            ┌──────────┐
            │ geometry │
            │ ---      │
            │ enum     │
            ╞══════════╡
            │ XY       │
            │ XYZ      │
            │ XYM      │
            │ XYZM     │
            └──────────┘
        """  # noqa: E501
        ...

    @register_plugin()
    def area(self) -> pl.Expr:
        """Return the area of each geometry."""
        ...

    @register_plugin()
    def bounds(self) -> pl.Expr:
        """Return the bounds of each geometry."""
        ...

    @register_plugin()
    def length(self) -> pl.Expr:
        """Return the length of each geometry."""
        ...

    @register_plugin()
    def minimum_clearance(self) -> pl.Expr:
        """Return the geometry minimum clearance."""
        ...

    @register_plugin()
    def x(self) -> pl.Expr:
        """Return the `x` value of Point geometries."""
        ...

    @register_plugin()
    def y(self) -> pl.Expr:
        """Return the `y` value of Point geometries."""
        ...

    @register_plugin()
    def z(self) -> pl.Expr:
        """Return the `z` value of Point geometries."""
        ...

    @register_plugin()
    def m(self) -> pl.Expr:
        """Return the `m` value of Point geometries."""
        ...

    @register_plugin()
    def count_coordinates(self) -> pl.Expr:
        """Return the number of coordinates in each geometry."""
        ...

    @register_plugin()
    def coordinates(self, output_dimension: CoordinateType | None = None) -> pl.Expr:
        """Return the coordinates of each geometry.

        Examples:
            >>> gdf = st.GeoDataFrame(["POINT(1 2 3)"])
            >>> gdf.select(st.geom().st.coordinates())
            shape: (1, 1)
            ┌───────────────────┐
            │ geometry          │
            │ ---               │
            │ list[list[f64]]   │
            ╞═══════════════════╡
            │ [[1.0, 2.0, 3.0]] │
            └───────────────────┘
        """
        ...

    @register_plugin()
    def exterior_ring(self) -> GeoExpr:
        """Return the exterior ring of Polygon geometries."""
        ...

    @register_plugin()
    def interior_rings(self) -> pl.Expr:
        """Return the list of interior rings for Polygon geometries."""
        ...

    @register_plugin()
    def count_interior_rings(self) -> pl.Expr:
        """Return the number of interior rings in Polygon geometries."""
        ...

    @register_plugin()
    def get_interior_ring(self, index: IntoIntegerExpr) -> GeoExpr:
        """Return the nth ring of Polygon geometries."""
        ...

    @register_plugin()
    def count_geometries(self) -> pl.Expr:
        """Return the number of parts in multipart geometries."""
        ...

    @register_plugin()
    def get_geometry(self, index: IntoIntegerExpr) -> GeoExpr:
        """Return the nth part of multipart geometries."""
        ...

    @register_plugin()
    def count_points(self) -> pl.Expr:
        """Return the number of points in LineString geometries."""
        ...

    @register_plugin()
    def get_point(self, index: IntoIntegerExpr) -> GeoExpr:
        """Return the nth point of LineString geometries."""
        ...

    @register_plugin()
    def parts(self) -> pl.Expr:
        """Return the list of parts for multipart geometries."""
        ...

    @register_plugin()
    def precision(self) -> pl.Expr:
        """Return the precision of each geometry."""
        ...

    @register_plugin()
    def set_precision(
        self,
        grid_size: IntoNumericExpr,
        mode: Literal["valid_output", "no_topo", "keep_collapsed"] = "valid_output",
    ) -> GeoExpr:
        """Set the precision of each geometry to a certain grid size."""
        ...

    @register_plugin()
    def distance(self, other: IntoGeoExprColumn) -> pl.Expr:
        """Return the distance from each geometry to other."""
        ...

    @register_plugin()
    def hausdorff_distance(
        self,
        other: IntoGeoExprColumn,
        densify: float | None = None,
    ) -> pl.Expr:
        """Return the hausdorff distance from each geometry to other."""
        ...

    @register_plugin()
    def frechet_distance(
        self,
        other: IntoGeoExprColumn,
        densify: float | None = None,
    ) -> pl.Expr:
        """Return the frechet distance from each geometry to other."""
        ...

    # Projection operations

    @register_plugin()
    def srid(self) -> pl.Expr:
        """Return the geometry SRID."""
        ...

    @register_plugin()
    def set_srid(self, srid: IntoIntegerExpr) -> GeoExpr:
        """Set the SRID of each geometry to a given value.

        Args:
            srid: The geometry new SRID
        """
        ...

    @register_plugin()
    def to_srid(self, srid: IntoIntegerExpr) -> GeoExpr:
        """Transform the coordinates of each geometry into a new CRS.

        Args:
            srid: The srid code of the new CRS

        Examples:
            >>> gdf = st.GeoDataFrame(["SRID=4326;POINT(0 0)"])
            >>> gdf = gdf.select(st.geom().st.to_srid(4301))
            >>> gdf.st.to_wkt()
            shape: (1, 1)
            ┌─────────────────────────────┐
            │ geometry                    │
            │ ---                         │
            │ str                         │
            ╞═════════════════════════════╡
            │ POINT (-0.004557 -0.006154) │
            └─────────────────────────────┘
        """
        ...

    # Serialization

    @register_plugin()
    def to_wkt(
        self,
        rounding_precision: int | None = 6,
        trim: bool = True,
        output_dimension: Literal[2, 3, 4] = 3,
        old_3d: bool = False,
    ) -> pl.Expr:
        """Serialize each geometry as WKT (Well-Known Text).

        Args:
            rounding_precision: The rounding precision when writing the WKT string.
                Set to None to indicate the full precision.
            trim: If True, trim unnecessary decimals (trailing zeros).
            output_dimension: The output dimension for the WKT string. Specifying 3
                means that up to 3 dimensions will be written but 2D geometries will
                still be represented as 2D in the WKT string.
            old_3d (bool, optional): Enable old style 3D/4D WKT generation. By default,
                new style 3D/4D WKT (ie. “POINT Z (10 20 30)”) is returned, but with
                `old_3d=True` the WKT will be formatted in the style “POINT (10 20 30)”.
        """
        ...

    @register_plugin()
    def to_ewkt(
        self,
        rounding_precision: int | None = 6,
        trim: bool = True,
        output_dimension: Literal[2, 3, 4] = 3,
        old_3d: bool = False,
    ) -> pl.Expr:
        """Serialize each geometry as EWKT (Extended Well-Known Text).

        Args:
            rounding_precision: The rounding precision when writing the WKT string.
                Set to None to indicate the full precision.
            trim: If True, trim unnecessary decimals (trailing zeros).
            output_dimension: The output dimension for the WKT string. Specifying 3
                means that up to 3 dimensions will be written but 2D geometries will
                still be represented as 2D in the WKT string.
            old_3d (bool, optional): Enable old style 3D/4D WKT generation. By default,
                new style 3D/4D WKT (ie. “POINT Z (10 20 30)”) is returned, but with
                `old_3d=True` the WKT will be formatted in the style “POINT (10 20 30)”.
        """
        ...

    @register_plugin()
    def to_wkb(
        self,
        output_dimension: Literal[2, 3, 4] = 3,
        byte_order: Literal[0, 1] | None = None,
        include_srid: bool = False,
    ) -> pl.Expr:
        """Serialize each geometry as WKB (Well-Known Binary).

        Args:
            output_dimension:
                The output dimension for the WKB. Specifying 3 means that up to 3 dimensions
                will be written but 2D geometries will still be represented as 2D in the WKB
                representation.
            byte_order:
                Defaults to native machine byte order (`None`). Use 0 to force big endian
                and 1 for little endian.
            include_srid:
                If True, the SRID is be included in WKB (this is an extension
                to the OGC WKB specification).
        """
        ...

    @register_plugin()
    def to_geojson(self, indent: int | None = None) -> pl.Expr:
        """Serialize each geometry as GeoJSON.

        Args:
            indent:
                If indent is not `None`, then GeoJSON will be pretty-printed.
                An indent level of 0 will only insert newlines. `None` (the default)
                outputs the most compact representation.
        """
        ...

    def to_shapely(self) -> pl.Expr:
        """Convert each geometry to a Shapely object."""
        import shapely

        return self._expr.map_batches(
            lambda s: pl.Series(s.name, shapely.from_wkb(s), dtype=pl.Object()),
            return_dtype=pl.Object(),
            is_elementwise=True,
        )

    def to_dict(self) -> pl.Expr:
        """Convert each geometry to a GeoJSON-like Python [`dict`][] object."""
        return self._expr.map_batches(
            lambda s: pl.Series(s.name, _lib.to_python_dict(s._s), dtype=pl.Object),  # noqa: SLF001
            return_dtype=pl.Object(),
            is_elementwise=True,
        )

    def cast(self, into: IntoExprColumn) -> pl.Expr:
        """Cast each geometry into a different compatible geometry type.

        Valid casts are:

        | Source          | Destination |
        |-----------------|-------------|
        | Point           | MultiPoint  |
        | MultiPoint      | LineString, CircularString |
        | LineString      | MultiPoint, CircularString, MultiLineString, MultiCurve |
        | CircularString  | MultiPoint, LineString, MultiLineString, MultiCurve |
        | MultiLineString | Polygon |
        | Polygon         | MultiPolygon, MultiSurface |
        | CurvePolygon    | MultiSurface |
        | Any             | GeometryCollection |
        """
        into = wrap_expr(parse_into_expression(into, str_as_lit=True))
        return register_plugin_function(
            plugin_path=Path(__file__).parent,
            function_name="cast",
            args=[self._expr, into],
            is_elementwise=True,
        )

    @register_plugin()
    def multi(self) -> pl.Expr:
        """Cast each geometry into their multipart equivalent."""
        ...

    # Unary predicates

    @register_plugin()
    def has_z(self) -> pl.Expr:
        """Return `True` for each geometry with `z` coordinate values."""
        ...

    @register_plugin()
    def has_m(self) -> pl.Expr:
        """Return `True` for each geometry with `m` coordinate values."""
        ...

    @register_plugin()
    def is_ccw(self) -> pl.Expr:
        """Return `True` for linear geometries with counter-clockwise coord sequence."""
        ...

    @register_plugin()
    def is_closed(self) -> pl.Expr:
        """Return `True` for closed linear geometries."""
        ...

    @register_plugin()
    def is_empty(self) -> pl.Expr:
        """Return `True` for empty geometries."""
        ...

    @register_plugin()
    def is_ring(self) -> pl.Expr:
        """Return `True` for ring geometries."""
        ...

    @register_plugin()
    def is_simple(self) -> pl.Expr:
        """Return `True` for simple geometries."""
        ...

    @register_plugin()
    def is_valid(self) -> pl.Expr:
        """Return `True` for valid geometries."""
        ...

    @register_plugin()
    def is_valid_reason(self) -> pl.Expr:
        """Return an explanation string for the invalidity of each geometry."""
        ...

    # Binary predicates

    @register_plugin()
    def crosses(self, other: IntoGeoExprColumn) -> pl.Expr:
        """Return `True` when each geometry crosses other."""
        ...

    @register_plugin()
    def contains(self, other: IntoGeoExprColumn) -> pl.Expr:
        """Return `True` when each geometry contains other."""
        ...

    @register_plugin()
    def contains_properly(self, other: IntoGeoExprColumn) -> pl.Expr:
        """Return `True` when each geometry properly contains other."""
        ...

    @register_plugin()
    def covered_by(self, other: IntoGeoExprColumn) -> pl.Expr:
        """Return `True` when each geometry is covered by other."""
        ...

    @register_plugin()
    def covers(self, other: IntoGeoExprColumn) -> pl.Expr:
        """Return `True` when each geometry covers other."""
        ...

    @register_plugin()
    def disjoint(self, other: IntoGeoExprColumn) -> pl.Expr:
        """Return `True` when each geometry is disjoint from other."""
        ...

    @register_plugin()
    def dwithin(self, other: IntoGeoExprColumn, distance: IntoNumericExpr) -> pl.Expr:
        """Return `True` when each geometry is within given distance to other."""
        ...

    @register_plugin()
    def intersects(self, other: IntoGeoExprColumn) -> pl.Expr:
        """Return `True` when each geometry intersects other."""
        ...

    @register_plugin()
    def overlaps(self, other: IntoGeoExprColumn) -> pl.Expr:
        """Return `True` when each geometry overlaps other."""
        ...

    @register_plugin()
    def touches(self, other: IntoGeoExprColumn) -> pl.Expr:
        """Return `True` when each geometry touches other."""
        ...

    @register_plugin()
    def within(self, other: IntoGeoExprColumn) -> pl.Expr:
        """Return `True` when each geometry is within other."""
        ...

    @register_plugin()
    def equals(self, other: IntoGeoExprColumn) -> pl.Expr:
        """Return `True` when each geometry is equal to other."""
        ...

    @register_plugin()
    def equals_exact(
        self,
        other: IntoGeoExprColumn,
        tolerance: float = 0.0,
    ) -> pl.Expr:
        """Return `True` when each geometry is equal to other."""
        ...

    @register_plugin()
    def equals_identical(self, other: IntoGeoExprColumn) -> pl.Expr:
        """Return `True` when each geometry is equal to other."""
        ...

    @register_plugin()
    def relate(self, other: IntoGeoExprColumn) -> pl.Expr:
        """Return the DE-9IM intersection matrix of each geometry with other."""
        ...

    @register_plugin()
    def relate_pattern(
        self,
        other: IntoGeoExprColumn,
        pattern: str,
    ) -> pl.Expr:
        """Return `True` when the DE-9IM intersection matrix of geometry with other matches a given pattern."""  # noqa: E501
        ...

    # Set operations

    @register_plugin()
    def union(self, other: IntoGeoExprColumn, grid_size: float | None = None) -> GeoExpr:
        """Return the union of each geometry with other."""
        ...

    @register_plugin()
    def unary_union(self, grid_size: float | None = None) -> GeoExpr:
        """Return the unary union of each geometry."""
        ...

    @register_plugin()
    def coverage_union(self) -> GeoExpr:
        """Return the coverage union of each geometry with other."""
        ...

    @register_plugin()
    def intersection(
        self,
        other: IntoGeoExprColumn,
        grid_size: float | None = None,
    ) -> GeoExpr:
        """Return the intersection of each geometry with other."""
        ...

    @register_plugin()
    def difference(
        self,
        other: IntoGeoExprColumn,
        grid_size: float | None = None,
    ) -> GeoExpr:
        """Return the difference of each geometry with other."""
        ...

    @register_plugin()
    def symmetric_difference(
        self,
        other: IntoGeoExprColumn,
        grid_size: float | None = None,
    ) -> GeoExpr:
        """Return the symmetric difference of each geometry with other."""
        ...

    # Constructive operations

    @register_plugin()
    def boundary(self) -> GeoExpr:
        """Return the topological boundary of each geometry."""
        ...

    @register_plugin()
    def buffer(
        self,
        distance: IntoNumericExpr,
        quad_segs: int = 8,
        cap_style: Literal["round", "square", "flat"] = "round",
        join_style: Literal["round", "mitre", "bevel"] = "round",
        mitre_limit: float = 5.0,
        single_sided: bool = False,
    ) -> GeoExpr:
        """Return a buffer around each geometry."""
        ...

    @register_plugin()
    def offset_curve(
        self,
        distance: IntoNumericExpr,
        quad_segs: int = 8,
        join_style: Literal["round", "mitre", "bevel"] = "round",
        mitre_limit: float = 5.0,
    ) -> GeoExpr:
        """Return a line at a given distance of each geometry."""
        ...

    @register_plugin()
    def centroid(self) -> GeoExpr:
        """Return the centroid of each geometry."""
        ...

    @register_plugin()
    def center(self) -> GeoExpr:
        """Return the bounding box center of each geometry."""
        ...

    @register_plugin()
    def clip_by_rect(self, bounds: IntoExprColumn) -> GeoExpr:
        """Clips each geometry by a bounding rectangle."""
        ...

    @register_plugin()
    def convex_hull(self) -> GeoExpr:
        """Return the convex hull of each geometry."""
        ...

    @register_plugin()
    def concave_hull(self, ratio: float = 0.0, allow_holes: bool = False) -> GeoExpr:
        """Return the concave hull of each geometry."""
        ...

    @register_plugin()
    def segmentize(self, max_segment_length: IntoNumericExpr) -> GeoExpr: ...

    @register_plugin()
    def envelope(self) -> GeoExpr:
        """Return the envelope of each geometry."""
        ...

    @register_plugin()
    def extract_unique_points(self) -> GeoExpr: ...

    @register_plugin()
    def build_area(self) -> GeoExpr: ...

    @register_plugin()
    def make_valid(
        self,
        method: Literal["linework", "structure"] = "linework",
        keep_collapsed: bool = True,
    ) -> GeoExpr: ...

    @register_plugin()
    def normalize(self) -> GeoExpr: ...

    @register_plugin()
    def node(self) -> GeoExpr: ...

    @register_plugin()
    def point_on_surface(self) -> GeoExpr:
        """Return a point that intersects of each geometry."""
        ...

    @register_plugin()
    def remove_repeated_points(self, tolerance: IntoNumericExpr = 0.0) -> GeoExpr:
        """Remove the repeated points for each geometry."""
        ...

    @register_plugin()
    def reverse(self) -> GeoExpr:
        """Reverse the coordinates order of each geometry."""
        ...

    @register_plugin()
    def simplify(
        self,
        tolerance: IntoNumericExpr,
        preserve_topology: bool = True,
    ) -> GeoExpr:
        """Simplify each geometry with a given tolerance."""
        ...

    @register_plugin()
    def force_2d(self) -> GeoExpr:
        """Force the dimensionality of a geometry to 2D."""
        ...

    @register_plugin()
    def force_3d(self, z: IntoNumericExpr = 0.0) -> GeoExpr:
        """Force the dimensionality of a geometry to 3D."""
        ...

    @register_plugin()
    def flip_coordinates(self) -> GeoExpr:
        """Flip the x and y coordinates of each geometry."""
        ...

    @register_plugin()
    def minimum_rotated_rectangle(self) -> GeoExpr: ...

    @register_plugin()
    def maximum_inscribed_circle(self, tolerance: IntoNumericExpr) -> GeoExpr:
        """Returns the maximum inscribed circle of a polygonal geometry."""
        ...

    @register_plugin()
    def snap(
        self,
        other: IntoGeoExprColumn,
        tolerance: IntoNumericExpr,
    ) -> GeoExpr: ...

    @register_plugin()
    def shortest_line(self, other: IntoGeoExprColumn) -> GeoExpr:
        """Return the shortest line between each geometry and other."""
        ...

    # Affine tranforms

    def affine_transform(self, matrix: IntoExprColumn | Sequence[float]) -> GeoExpr:
        """Apply a 2D or 3D transformation matrix to the coordinates of each geometry.

        Args:
            matrix:
                The transformation matrix to apply to coordinates. Should contains 6
                elements for a 2D transform or 12 for a 3D transform. The matrix elements
                order should be, in order:
                - `m11`, `m12`, `m21`, `m22`, `tx`, `ty` for 2D transformations
                - `m11`, `m12`, `m13`, `m21`, `m22`, `m23`, `m31`, `m32`, `m33`, `tx`, `ty`, `tz`
                    for 3D transformations

        """
        return register_plugin_function(
            plugin_path=Path(__file__).parent,
            function_name="affine_transform",
            args=[
                self._expr,
                matrix
                if isinstance(matrix, pl.Expr | pl.Series | str)
                else pl.lit(matrix, dtype=pl.Array(pl.Float64, len(matrix))),
            ],
            is_elementwise=True,
        ).pipe(lambda e: cast("GeoExpr", e))

    def translate(
        self,
        x: IntoNumericExpr = 0.0,
        y: IntoNumericExpr = 0.0,
        z: IntoNumericExpr = 0.0,
    ) -> GeoExpr:
        return register_plugin_function(
            plugin_path=Path(__file__).parent,
            function_name="translate",
            args=[self._expr, pl.concat_list(x, y, z)],
            is_elementwise=True,
        ).pipe(lambda e: cast("GeoExpr", e))

    @register_plugin()
    def rotate(
        self,
        angle: IntoNumericExpr,
        origin: Literal["center", "centroid"] | Sequence[float] = "center",
    ) -> GeoExpr: ...

    def scale(
        self,
        x: IntoNumericExpr = 1.0,
        y: IntoNumericExpr = 1.0,
        z: IntoNumericExpr = 1.0,
        origin: Literal["center", "centroid"] | Sequence[float] = "center",
    ) -> GeoExpr:
        return register_plugin_function(
            plugin_path=Path(__file__).parent,
            function_name="scale",
            args=[self._expr, pl.concat_list(x, y, z)],
            kwargs={"origin": origin},
            is_elementwise=True,
        ).pipe(lambda e: cast("GeoExpr", e))

    def skew(
        self,
        x: IntoNumericExpr = 0.0,
        y: IntoNumericExpr = 0.0,
        z: IntoNumericExpr = 0.0,
        origin: Literal["center", "centroid"] | Sequence[float] = "center",
    ) -> GeoExpr:
        return register_plugin_function(
            plugin_path=Path(__file__).parent,
            function_name="skew",
            args=[self._expr, pl.concat_list(x, y, z)],
            kwargs={"origin": origin},
            is_elementwise=True,
        ).pipe(lambda e: cast("GeoExpr", e))

    # Linestring operations

    @register_plugin()
    def interpolate(
        self,
        distance: IntoNumericExpr,
        normalized: bool = False,
    ) -> GeoExpr: ...

    @register_plugin()
    def project(
        self,
        other: IntoGeoExprColumn,
        normalized: bool = False,
    ) -> pl.Expr: ...

    @register_plugin()
    def substring(self, start: IntoNumericExpr, end: IntoNumericExpr) -> GeoExpr:
        """Returns the substring of each line starting and ending at the given fractional locations."""  # noqa: E501
        ...

    @register_plugin()
    def line_merge(self, directed: bool = False) -> GeoExpr: ...

    @register_plugin()
    def shared_paths(self, other: IntoGeoExprColumn) -> GeoExpr: ...

    # Aggregations

    @register_plugin(is_aggregation=True)
    def total_bounds(self) -> pl.Expr:
        """Return the total bounds of all geometries."""
        ...

    @register_plugin(is_aggregation=True)
    def collect(self, into: GeometryType | None = None) -> GeoExpr:
        """Aggregate geometries into a single collection."""
        ...

    @register_plugin(is_aggregation=True)
    def union_all(self, grid_size: float | None = None) -> GeoExpr:
        """Return the union of all geometries."""
        ...

    @register_plugin(is_aggregation=True)
    def coverage_union_all(self) -> GeoExpr:
        """Return the coverage union of all geometries."""
        ...

    @register_plugin(is_aggregation=True)
    def intersection_all(self, grid_size: float | None = None) -> GeoExpr:
        """Return the intersection of all geometries."""
        ...

    @register_plugin(is_aggregation=True)
    def difference_all(self, grid_size: float | None = None) -> GeoExpr:
        """Return the difference of all geometries."""
        ...

    @register_plugin(is_aggregation=True)
    def symmetric_difference_all(self, grid_size: float | None = None) -> GeoExpr:
        """Return the symmetric difference of all geometries."""
        ...

    @register_plugin(is_aggregation=True)
    def polygonize(self) -> GeoExpr: ...

    @register_plugin(is_aggregation=True)
    def voronoi_polygons(
        self,
        tolerance: float = 0.0,
        extend_to: bytes | None = None,
        only_edges: bool = False,
    ) -> GeoExpr:
        """Return a Voronoi diagram of all geometries vertices."""
        ...

    @register_plugin(is_aggregation=True)
    def delaunay_triangles(
        self,
        tolerance: float = 0.0,
        only_edges: bool = False,
    ) -> GeoExpr:
        """Return a Delaunay triangulation of all geometries vertices."""
        ...



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from polars_st.selectors import geom

if TYPE_CHECKING:
    from collections.abc import Sequence

    import polars as pl

    from polars_st.geometry import CoordinateType, GeometryType
    from polars_st.typing import (
        IntoExprColumn,
        IntoIntegerExpr,
        IntoNumericExpr,
    )

    from .geoexpr import GeoExpr


__all__ = [
    "affine_transform",
    "area",
    "boundary",
    "bounds",
    "buffer",
    "build_area",
    "cast",
    "center",
    "centroid",
    "clip_by_rect",
    "collect",
    "concave_hull",
    "convex_hull",
    "coordinate_dimension",
    "coordinate_type",
    "coordinates",
    "count_coordinates",
    "count_geometries",
    "count_interior_rings",
    "count_points",
    "coverage_union",
    "coverage_union_all",
    "delaunay_triangles",
    "difference_all",
    "dimension",
    "envelope",
    "exterior_ring",
    "extract_unique_points",
    "flip_coordinates",
    "force_2d",
    "force_3d",
    "geometry_type",
    "get_geometry",
    "get_interior_ring",
    "get_point",
    "has_m",
    "has_z",
    "interior_rings",
    "interpolate",
    "intersection_all",
    "is_ccw",
    "is_closed",
    "is_empty",
    "is_ring",
    "is_simple",
    "is_valid",
    "is_valid_reason",
    "length",
    "line_merge",
    "m",
    "make_valid",
    "maximum_inscribed_circle",
    "minimum_clearance",
    "minimum_rotated_rectangle",
    "multi",
    "node",
    "normalize",
    "offset_curve",
    "parts",
    "point_on_surface",
    "polygonize",
    "precision",
    "remove_repeated_points",
    "reverse",
    "rotate",
    "scale",
    "segmentize",
    "set_precision",
    "set_srid",
    "simplify",
    "skew",
    "srid",
    "substring",
    "symmetric_difference_all",
    "to_dict",
    "to_ewkt",
    "to_geojson",
    "to_shapely",
    "to_srid",
    "to_wkb",
    "to_wkt",
    "total_bounds",
    "translate",
    "unary_union",
    "union_all",
    "voronoi_polygons",
    "x",
    "y",
    "z",
]


def geometry_type(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[geometry_type()][polars_st.GeoExprNameSpace.geometry_type]</code>."""  # noqa: E501
    return geom(*columns).st.geometry_type()


def dimension(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[dimension()][polars_st.GeoExprNameSpace.dimension]</code>."""  # noqa: E501
    return geom(*columns).st.dimension()


def coordinate_dimension(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[coordinate_dimension()][polars_st.GeoExprNameSpace.coordinate_dimension]</code>."""  # noqa: E501
    return geom(*columns).st.coordinate_dimension()


def coordinate_type(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[coordinate_type()][polars_st.GeoExprNameSpace.coordinate_type]</code>."""  # noqa: E501
    return geom(*columns).st.coordinate_type()


def area(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[area()][polars_st.GeoExprNameSpace.area]</code>."""  # noqa: E501
    return geom(*columns).st.area()


def bounds(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[bounds()][polars_st.GeoExprNameSpace.bounds]</code>."""  # noqa: E501
    return geom(*columns).st.bounds()


def length(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[length()][polars_st.GeoExprNameSpace.length]</code>."""  # noqa: E501
    return geom(*columns).st.length()


def minimum_clearance(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[minimum_clearance()][polars_st.GeoExprNameSpace.minimum_clearance]</code>."""  # noqa: E501
    return geom(*columns).st.minimum_clearance()


def x(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[x()][polars_st.GeoExprNameSpace.x]</code>."""
    return geom(*columns).st.x()


def y(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[y()][polars_st.GeoExprNameSpace.y]</code>."""
    return geom(*columns).st.y()


def z(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[z()][polars_st.GeoExprNameSpace.z]</code>."""
    return geom(*columns).st.z()


def m(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[m()][polars_st.GeoExprNameSpace.m]</code>."""
    return geom(*columns).st.m()


def count_coordinates(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[count_coordinates()][polars_st.GeoExprNameSpace.count_coordinates]</code>."""  # noqa: E501
    return geom(*columns).st.count_coordinates()


def coordinates(*columns: str, output_dimension: CoordinateType | None = None) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[coordinates(...)][polars_st.GeoExprNameSpace.coordinates]</code>."""  # noqa: E501
    return geom(*columns).st.coordinates(output_dimension)


def exterior_ring(*columns: str) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[exterior_ring()][polars_st.GeoExprNameSpace.exterior_ring]</code>."""  # noqa: E501
    return geom(*columns).st.exterior_ring()


def interior_rings(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[interior_rings()][polars_st.GeoExprNameSpace.interior_rings]</code>."""  # noqa: E501
    return geom(*columns).st.interior_rings()


def count_interior_rings(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[count_interior_rings()][polars_st.GeoExprNameSpace.count_interior_rings]</code>."""  # noqa: E501
    return geom(*columns).st.count_interior_rings()


def get_interior_ring(*columns: str, index: IntoIntegerExpr) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[get_interior_ring(...)][polars_st.GeoExprNameSpace.get_interior_ring]</code>."""  # noqa: E501
    return geom(*columns).st.get_interior_ring(index)


def count_geometries(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[count_geometries()][polars_st.GeoExprNameSpace.count_geometries]</code>."""  # noqa: E501
    return geom(*columns).st.count_geometries()


def get_geometry(*columns: str, index: IntoIntegerExpr) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[get_geometry(...)][polars_st.GeoExprNameSpace.get_geometry]</code>."""  # noqa: E501
    return geom(*columns).st.get_geometry(index)


def count_points(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[count_points()][polars_st.GeoExprNameSpace.count_points]</code>."""  # noqa: E501
    return geom(*columns).st.count_points()


def get_point(*columns: str, index: IntoIntegerExpr) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[get_point(...)][polars_st.GeoExprNameSpace.get_point]</code>."""  # noqa: E501
    return geom(*columns).st.get_point(index)


def parts(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[parts()][polars_st.GeoExprNameSpace.parts]</code>."""  # noqa: E501
    return geom(*columns).st.parts()


def precision(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[precision()][polars_st.GeoExprNameSpace.precision]</code>."""  # noqa: E501
    return geom(*columns).st.precision()


def set_precision(
    *columns: str,
    grid_size: IntoNumericExpr,
    mode: Literal["valid_output", "no_topo", "keep_collapsed"] = "valid_output",
) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[set_precision(...)][polars_st.GeoExprNameSpace.set_precision]</code>."""  # noqa: E501
    return geom(*columns).st.set_precision(grid_size, mode)


def srid(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[srid()][polars_st.GeoExprNameSpace.srid]</code>."""  # noqa: E501
    return geom(*columns).st.srid()


def set_srid(*columns: str, srid: IntoIntegerExpr) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[set_srid(...)][polars_st.GeoExprNameSpace.set_srid]</code>."""  # noqa: E501
    return geom(*columns).st.set_srid(srid)


def to_srid(*columns: str, srid: IntoIntegerExpr) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[to_srid(...)][polars_st.GeoExprNameSpace.to_srid]</code>."""  # noqa: E501
    return geom(*columns).st.to_srid(srid)


def to_wkt(
    *columns: str,
    rounding_precision: int | None = 6,
    trim: bool = True,
    output_dimension: Literal[2, 3, 4] = 3,
    old_3d: bool = False,
) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[to_wkt(...)][polars_st.GeoExprNameSpace.to_wkt]</code>."""  # noqa: E501
    return geom(*columns).st.to_wkt(rounding_precision, trim, output_dimension, old_3d)


def to_ewkt(
    *columns: str,
    rounding_precision: int | None = 6,
    trim: bool = True,
    output_dimension: Literal[2, 3, 4] = 3,
    old_3d: bool = False,
) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[to_ewkt(...)][polars_st.GeoExprNameSpace.to_ewkt]</code>."""  # noqa: E501
    return geom(*columns).st.to_ewkt(rounding_precision, trim, output_dimension, old_3d)


def to_wkb(
    *columns: str,
    output_dimension: Literal[2, 3, 4] = 3,
    byte_order: Literal[0, 1] | None = None,
    include_srid: bool = False,
) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[to_wkb(...)][polars_st.GeoExprNameSpace.to_wkb]</code>."""  # noqa: E501
    return geom(*columns).st.to_wkb(output_dimension, byte_order, include_srid)


def to_geojson(*columns: str, indent: int | None = None) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[to_geojson(...)][polars_st.GeoExprNameSpace.to_geojson]</code>."""  # noqa: E501
    return geom(*columns).st.to_geojson(indent)


def to_shapely(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[to_shapely()][polars_st.GeoExprNameSpace.to_shapely]</code>."""  # noqa: E501
    return geom(*columns).st.to_shapely()


def to_dict(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[to_dict()][polars_st.GeoExprNameSpace.to_dict]</code>."""  # noqa: E501
    return geom(*columns).st.to_dict()


def cast(*columns: str, into: IntoExprColumn) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[cast(into)][polars_st.GeoExprNameSpace.cast]</code>."""  # noqa: E501
    return geom(*columns).st.cast(into)


def multi(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[multi()][polars_st.GeoExprNameSpace.multi]</code>."""  # noqa: E501
    return geom(*columns).st.multi()


def has_z(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[has_z()][polars_st.GeoExprNameSpace.has_z]</code>."""  # noqa: E501
    return geom(*columns).st.has_z()


def has_m(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[has_m()][polars_st.GeoExprNameSpace.has_m]</code>."""  # noqa: E501
    return geom(*columns).st.has_m()


def is_ccw(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[is_ccw()][polars_st.GeoExprNameSpace.is_ccw]</code>."""  # noqa: E501
    return geom(*columns).st.is_ccw()


def is_closed(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[is_closed()][polars_st.GeoExprNameSpace.is_closed]</code>."""  # noqa: E501
    return geom(*columns).st.is_closed()


def is_empty(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[is_empty()][polars_st.GeoExprNameSpace.is_empty]</code>."""  # noqa: E501
    return geom(*columns).st.is_empty()


def is_ring(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[is_ring()][polars_st.GeoExprNameSpace.is_ring]</code>."""  # noqa: E501
    return geom(*columns).st.is_ring()


def is_simple(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[is_simple()][polars_st.GeoExprNameSpace.is_simple]</code>."""  # noqa: E501
    return geom(*columns).st.is_simple()


def is_valid(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[is_valid()][polars_st.GeoExprNameSpace.is_valid]</code>."""  # noqa: E501
    return geom(*columns).st.is_valid()


def is_valid_reason(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[is_valid_reason()][polars_st.GeoExprNameSpace.is_valid_reason]</code>."""  # noqa: E501
    return geom(*columns).st.is_valid_reason()


def unary_union(*columns: str, grid_size: float | None = None) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[unary_union(...)][polars_st.GeoExprNameSpace.unary_union]</code>."""  # noqa: E501
    return geom(*columns).st.unary_union(grid_size)


def coverage_union(*columns: str) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[coverage_union()][polars_st.GeoExprNameSpace.coverage_union]</code>."""  # noqa: E501
    return geom(*columns).st.coverage_union()


def boundary(*columns: str) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[boundary()][polars_st.GeoExprNameSpace.boundary]</code>."""  # noqa: E501
    return geom(*columns).st.boundary()


def buffer(
    *columns: str,
    distance: IntoNumericExpr,
    quad_segs: int = 8,
    cap_style: Literal["round", "square", "flat"] = "round",
    join_style: Literal["round", "mitre", "bevel"] = "round",
    mitre_limit: float = 5.0,
    single_sided: bool = False,
) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[buffer(...)][polars_st.GeoExprNameSpace.buffer]</code>."""  # noqa: E501
    return geom(*columns).st.buffer(
        distance,
        quad_segs,
        cap_style,
        join_style,
        mitre_limit,
        single_sided,
    )


def offset_curve(
    *columns: str,
    distance: IntoNumericExpr,
    quad_segs: int = 8,
    join_style: Literal["round", "mitre", "bevel"] = "round",
    mitre_limit: float = 5.0,
) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[offset_curve(...)][polars_st.GeoExprNameSpace.offset_curve]</code>."""  # noqa: E501
    return geom(*columns).st.offset_curve(distance, quad_segs, join_style, mitre_limit)


def centroid(*columns: str) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[centroid()][polars_st.GeoExprNameSpace.centroid]</code>."""  # noqa: E501
    return geom(*columns).st.centroid()


def center(*columns: str) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[center()][polars_st.GeoExprNameSpace.center]</code>."""  # noqa: E501
    return geom(*columns).st.center()


def clip_by_rect(*columns: str, bounds: IntoExprColumn) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[clip_by_rect(...)][polars_st.GeoExprNameSpace.clip_by_rect]</code>."""  # noqa: E501
    return geom(*columns).st.clip_by_rect(bounds)


def convex_hull(*columns: str) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[convex_hull()][polars_st.GeoExprNameSpace.convex_hull]</code>."""  # noqa: E501
    return geom(*columns).st.convex_hull()


def concave_hull(*columns: str, ratio: float = 0.0, allow_holes: bool = False) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[concave_hull(...)][polars_st.GeoExprNameSpace.concave_hull]</code>."""  # noqa: E501
    return geom(*columns).st.concave_hull(ratio, allow_holes)


def segmentize(*columns: str, max_segment_length: IntoNumericExpr) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[segmentize(max_segment_length)][polars_st.GeoExprNameSpace.segmentize]</code>."""  # noqa: E501
    return geom(*columns).st.segmentize(max_segment_length)


def envelope(*columns: str) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[envelope()][polars_st.GeoExprNameSpace.envelope]</code>."""  # noqa: E501
    return geom(*columns).st.envelope()


def extract_unique_points(*columns: str) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[extract_unique_points()][polars_st.GeoExprNameSpace.extract_unique_points]</code>."""  # noqa: E501
    return geom(*columns).st.extract_unique_points()


def build_area(*columns: str) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[build_area()][polars_st.GeoExprNameSpace.build_area]</code>."""  # noqa: E501
    return geom(*columns).st.build_area()


def make_valid(
    *columns: str,
    method: Literal["linework", "structure"] = "linework",
    keep_collapsed: bool = True,
) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[make_valid()][polars_st.GeoExprNameSpace.make_valid]</code>."""  # noqa: E501
    return geom(*columns).st.make_valid(method, keep_collapsed)


def normalize(*columns: str) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[normalize()][polars_st.GeoExprNameSpace.normalize]</code>."""  # noqa: E501
    return geom(*columns).st.normalize()


def node(*columns: str) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[node()][polars_st.GeoExprNameSpace.node]</code>."""  # noqa: E501
    return geom(*columns).st.node()


def point_on_surface(*columns: str) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[point_on_surface()][polars_st.GeoExprNameSpace.point_on_surface]</code>."""  # noqa: E501
    return geom(*columns).st.point_on_surface()


def remove_repeated_points(*columns: str, tolerance: IntoNumericExpr = 0.0) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[remove_repeated_points(...)][polars_st.GeoExprNameSpace.remove_repeated_points]</code>."""  # noqa: E501
    return geom(*columns).st.remove_repeated_points(tolerance)


def reverse(*columns: str) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[reverse()][polars_st.GeoExprNameSpace.reverse]</code>."""  # noqa: E501
    return geom(*columns).st.reverse()


def simplify(
    *columns: str,
    tolerance: IntoNumericExpr,
    preserve_topology: bool = True,
) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[simplify(...)][polars_st.GeoExprNameSpace.simplify]</code>."""  # noqa: E501
    return geom(*columns).st.simplify(tolerance, preserve_topology)


def force_2d(*columns: str) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[force_2d()][polars_st.GeoExprNameSpace.force_2d]</code>."""  # noqa: E501
    return geom(*columns).st.force_2d()


def force_3d(*columns: str, z: IntoNumericExpr = 0.0) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[force_3d(...)][polars_st.GeoExprNameSpace.force_3d]</code>."""  # noqa: E501
    return geom(*columns).st.force_3d(z)


def flip_coordinates(*columns: str) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[flip_coordinates()][polars_st.GeoExprNameSpace.flip_coordinates]</code>."""  # noqa: E501
    return geom(*columns).st.flip_coordinates()


def minimum_rotated_rectangle(*columns: str) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[minimum_rotated_rectangle()][polars_st.GeoExprNameSpace.minimum_rotated_rectangle]</code>."""  # noqa: E501
    return geom(*columns).st.minimum_rotated_rectangle()


def maximum_inscribed_circle(*columns: str, tolerance: IntoNumericExpr) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[maximum_inscribed_circle(...)][polars_st.GeoExprNameSpace.maximum_inscribed_circle]</code>."""  # noqa: E501
    return geom(*columns).st.maximum_inscribed_circle(tolerance)


def affine_transform(*columns: str, matrix: IntoExprColumn | Sequence[float]) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[affine_transform(...)][polars_st.GeoExprNameSpace.affine_transform]</code>."""  # noqa: E501
    return geom(*columns).st.affine_transform(matrix)


def translate(
    *columns: str,
    x: IntoNumericExpr = 0.0,
    y: IntoNumericExpr = 0.0,
    z: IntoNumericExpr = 0.0,
) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[translate(...)][polars_st.GeoExprNameSpace.translate]</code>."""  # noqa: E501
    return geom(*columns).st.translate(x, y, z)


def rotate(
    *columns: str,
    angle: IntoNumericExpr,
    origin: Literal["center", "centroid"] | Sequence[float] = "center",
) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[rotate(...)][polars_st.GeoExprNameSpace.rotate]</code>."""  # noqa: E501
    return geom(*columns).st.rotate(angle, origin)


def scale(
    *columns: str,
    x: IntoNumericExpr = 1.0,
    y: IntoNumericExpr = 1.0,
    z: IntoNumericExpr = 1.0,
    origin: Literal["center", "centroid"] | Sequence[float] = "center",
) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[scale(...)][polars_st.GeoExprNameSpace.scale]</code>."""  # noqa: E501
    return geom(*columns).st.scale(x, y, z, origin)


def skew(
    *columns: str,
    x: IntoNumericExpr = 0.0,
    y: IntoNumericExpr = 0.0,
    z: IntoNumericExpr = 0.0,
    origin: Literal["center", "centroid"] | Sequence[float] = "center",
) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[skew(...)][polars_st.GeoExprNameSpace.skew]</code>."""  # noqa: E501
    return geom(*columns).st.skew(x, y, z, origin)


def interpolate(
    *columns: str,
    distance: IntoNumericExpr,
    normalized: bool = False,
) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[interpolate()][polars_st.GeoExprNameSpace.interpolate]</code>."""  # noqa: E501
    return geom(*columns).st.interpolate(distance, normalized)


def substring(
    *columns: str,
    start: IntoNumericExpr,
    end: IntoNumericExpr,
) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[substring(...)][polars_st.GeoExprNameSpace.substring]</code>."""  # noqa: E501
    return geom(*columns).st.substring(start, end)


def line_merge(*columns: str, directed: bool = False) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[line_merge()][polars_st.GeoExprNameSpace.line_merge]</code>."""  # noqa: E501
    return geom(*columns).st.line_merge(directed)


def total_bounds(*columns: str) -> pl.Expr:
    """Syntactic sugar for <code>st.geom(columns).st.[total_bounds()][polars_st.GeoExprNameSpace.total_bounds]</code>."""  # noqa: E501
    return geom(*columns).st.total_bounds()


def collect(*columns: str, into: GeometryType | None = None) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[collect()][polars_st.GeoExprNameSpace.collect]</code>."""  # noqa: E501
    return geom(*columns).st.collect(into)


def union_all(*columns: str, grid_size: float | None = None) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[union_all(...)][polars_st.GeoExprNameSpace.union_all]</code>."""  # noqa: E501
    return geom(*columns).st.union_all(grid_size)


def coverage_union_all(*columns: str) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[coverage_union_all()][polars_st.GeoExprNameSpace.coverage_union_all]</code>."""  # noqa: E501
    return geom(*columns).st.coverage_union_all()


def intersection_all(*columns: str, grid_size: float | None = None) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[intersection_all()][polars_st.GeoExprNameSpace.intersection_all]</code>."""  # noqa: E501
    return geom(*columns).st.intersection_all(grid_size)


def difference_all(*columns: str, grid_size: float | None = None) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[difference_all(...)][polars_st.GeoExprNameSpace.difference_all]</code>."""  # noqa: E501
    return geom(*columns).st.difference_all(grid_size)


def symmetric_difference_all(*columns: str, grid_size: float | None = None) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[symmetric_difference_all(...)][polars_st.GeoExprNameSpace.symmetric_difference_all]</code>."""  # noqa: E501
    return geom(*columns).st.symmetric_difference_all(grid_size)


def polygonize(*columns: str) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[polygonize()][polars_st.GeoExprNameSpace.polygonize]</code>."""  # noqa: E501
    return geom(*columns).st.polygonize()


def voronoi_polygons(
    *columns: str,
    tolerance: float = 0.0,
    extend_to: bytes | None = None,
    only_edges: bool = False,
) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[voronoi_polygons(...)][polars_st.GeoExprNameSpace.voronoi_polygons]</code>."""  # noqa: E501
    return geom(*columns).st.voronoi_polygons(tolerance, extend_to, only_edges)


def delaunay_triangles(
    *columns: str,
    tolerance: float = 0.0,
    only_edges: bool = False,
) -> GeoExpr:
    """Syntactic sugar for <code>st.geom(columns).st.[delaunay_triangles(...)][polars_st.GeoExprNameSpace.delaunay_triangles]</code>."""  # noqa: E501
    return geom(*columns).st.delaunay_triangles(tolerance, only_edges)
